"""Local frequency analysis functions and utilities."""

import datetime
from typing import Optional, Union

import numpy as np
import statsmodels
import xarray as xr
import xclim.indices.stats
from scipy.stats.mstats import plotting_positions
from statsmodels.tools import eval_measures

__all__ = [
    "criteria",
    "fit",
    "parametric_quantiles",
]


def fit(
    ds,
    distributions: Optional[list[str]] = None,
    min_years: Optional[int] = None,
    method: str = "ML",
) -> xr.Dataset:
    """Fit multiple distributions to data.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset containing the data to fit. All variables will be fitted.
    distributions : list of str, optional
        List of distribution names as defined in `scipy.stats`. See https://docs.scipy.org/doc/scipy/reference/stats.html#continuous-distributions.
        Defaults to ["expon", "gamma", "genextreme", "genpareto", "gumbel_r", "pearson3", "weibull_min"].
    min_years : int, optional
        Minimum number of years required for a distribution to be fitted.
    method : str
        Fitting method. Defaults to "ML" (maximum likelihood).

    Returns
    -------
    xr.Dataset
        Dataset containing the parameters of the fitted distributions, with a new dimension `scipy_dist` containing the distribution names.

    Notes
    -----
    In order to combine the parameters of multiple distributions, the size of the `dparams` dimension is set to the
    maximum number of unique parameters between the distributions.
    """
    distributions = distributions or [
        "expon",
        "gamma",
        "genextreme",
        "genpareto",
        "gumbel_r",
        "pearson3",
        "weibull_min",
    ]
    out = []
    for v in ds.data_vars:
        p = []
        for d in distributions:
            p.append(
                xclim.indices.stats.fit(
                    ds[v].chunk({"time": -1}), dist=d, method=method
                )
                .assign_coords(scipy_dist=d)
                .expand_dims("scipy_dist")
            )
        params = xr.concat(p, dim="scipy_dist")

        # Reorder dparams to match the order of the parameters across all distributions, since subsequent operations rely on this.
        p_order = sorted(set(params.dparams.values).difference(["loc", "scale"])) + [
            "loc",
            "scale",
        ]
        params = params.sel(dparams=p_order)

        if min_years is not None:
            params = params.where(ds[v].notnull().sum("time") >= min_years)
        params.attrs["scipy_dist"] = distributions
        params.attrs["description"] = "Parameters of the distributions"
        params.attrs["long_name"] = "Distribution parameters"
        params.attrs["min_years"] = min_years
        out.append(params)

    out = xr.merge(out)
    out.attrs = ds.attrs

    return out


def parametric_quantiles(
    p: xr.Dataset, t: Union[float, list[float]], mode: str = "max"
) -> xr.Dataset:
    """Compute quantiles from fitted distributions.

    Parameters
    ----------
    p : xr.Dataset
        Dataset containing the parameters of the fitted distributions.
        Must have a dimension `dparams` containing the parameter names and a dimension `scipy_dist` containing the distribution names.
    t : float or list of float
        Return period(s) in years.
    mode : {'max', 'min'}
        Whether the return period is the probability of exceedance (max) or non-exceedance (min).

    Returns
    -------
    xr.Dataset
        Dataset containing the quantiles of the distributions.
    """
    distributions = list(p["scipy_dist"].values)

    t = np.atleast_1d(t)
    if mode == "max":
        q = 1 - 1.0 / t
    elif mode == "min":
        q = 1.0 / t
    else:
        raise ValueError(f"'mode' must be 'max' or 'min', got '{mode}'.")

    out = []
    for v in p.data_vars:
        quantiles = []
        for d in distributions:
            dist_obj = xclim.indices.stats.get_dist(d)
            shape_params = [] if dist_obj.shapes is None else dist_obj.shapes.split(",")
            dist_params = shape_params + ["loc", "scale"]
            da = p[v].sel(scipy_dist=d, dparams=dist_params).transpose("dparams", ...)
            da.attrs["scipy_dist"] = d
            qt = (
                xclim.indices.stats.parametric_quantile(da, q=q)
                .rename({"quantile": "return_period"})
                .assign_coords(scipy_dist=d, return_period=t)
                .expand_dims("scipy_dist")
            )
            quantiles.append(qt)
        quantiles = xr.concat(quantiles, dim="scipy_dist")

        # Add the quantile as a new coordinate
        da_q = xr.DataArray(q, dims="return_period", coords={"return_period": t})
        da_q.attrs["long_name"] = (
            "Probability of exceedance"
            if mode == "max"
            else "Probability of non-exceedance"
        )
        da_q.attrs["description"] = (
            "Parametric distribution quantiles for the given return period."
        )
        da_q.attrs["mode"] = mode
        quantiles = quantiles.assign_coords(p_quantile=da_q)

        quantiles.attrs["scipy_dist"] = distributions
        quantiles.attrs["description"] = (
            f"Return period ({mode}) estimated with statistic distributions"
        )
        quantiles.attrs["long_name"] = "Return period"
        quantiles.attrs["mode"] = mode
        out.append(quantiles)

    out = xr.merge(out)
    out.attrs = p.attrs

    return out


def criteria(ds: xr.Dataset, p: xr.Dataset) -> xr.Dataset:
    """Compute information criteria (AIC, BIC, AICC) from fitted distributions, using the log-likelihood.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset containing the yearly data that was fitted.
    p : xr.Dataset
        Dataset containing the parameters of the fitted distributions.
        Must have a dimension `dparams` containing the parameter names and a dimension `scipy_dist` containing the distribution names.

    Returns
    -------
    xr.Dataset
        Dataset containing the information criteria for the distributions.
    """

    def _get_criteria_1d(da, params, dist):
        da = np.atleast_2d(da).T
        params = np.atleast_2d(params).T

        llf = np.nansum(dist.logpdf(da, *params), axis=0)  # log-likelihood
        nobs = np.sum(np.isfinite(da), axis=0)
        df_modelwc = len(p)
        dof_eff = nobs - df_modelwc - 1.0

        aic = eval_measures.aic(llf=llf, nobs=nobs, df_modelwc=len(p))
        bic = eval_measures.bic(llf=llf, nobs=nobs, df_modelwc=len(p))
        # Custom AICC, because the one in statsmodels does not support multiple dimensions
        aicc = np.where(
            dof_eff > 0, -2.0 * llf + 2.0 * df_modelwc * nobs / dof_eff, np.nan
        )

        return np.stack([aic, bic, aicc], axis=1)

    out = []
    distributions = list(p["scipy_dist"].values)

    common_vars = list(set(ds.data_vars).intersection(p.data_vars))
    for v in common_vars:
        c = []
        for d in distributions:
            dist_obj = xclim.indices.stats.get_dist(d)
            shape_params = [] if dist_obj.shapes is None else dist_obj.shapes.split(",")
            dist_params = shape_params + ["loc", "scale"]
            da = ds[v].transpose("time", ...)
            params = (
                p[v].sel(scipy_dist=d, dparams=dist_params).transpose("dparams", ...)
            )

            if len(da.dims) == 1:
                crit = xr.apply_ufunc(
                    _get_criteria_1d,
                    da.expand_dims("tmp"),
                    params.expand_dims("tmp"),
                    kwargs={"dist": dist_obj},
                    input_core_dims=[["time"], ["dparams"]],
                    output_core_dims=[["criterion"]],
                    dask_gufunc_kwargs=dict(output_sizes={"criterion": 3}),
                    dask="parallelized",
                ).squeeze("tmp")
            else:
                crit = xr.apply_ufunc(
                    _get_criteria_1d,
                    da,
                    params,
                    kwargs={"dist": dist_obj},
                    input_core_dims=[["time"], ["dparams"]],
                    output_core_dims=[["criterion"]],
                    dask_gufunc_kwargs=dict(output_sizes={"criterion": 3}),
                    dask="parallelized",
                )

            # Add attributes
            crit = crit.assign_coords(criterion=["aic", "bic", "aicc"]).expand_dims(
                "scipy_dist"
            )
            crit.attrs = p[v].attrs
            crit.attrs["history"] = (
                crit.attrs.get("history", "")
                + f", [{datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] "
                f"criteria: computed AIC, BIC and AICC. - statsmodels version: {statsmodels.__version__}"
            )
            crit.attrs["description"] = (
                "Information criteria for the distribution parameters."
            )
            crit.attrs["long_name"] = "Information criteria"

            # Remove a few attributes that are not relevant anymore
            crit.attrs.pop("method", None)
            crit.attrs.pop("estimator", None)
            crit.attrs.pop("min_years", None)
            c.append(crit)

        c = xr.concat(c, dim="scipy_dist")
        out.append(c)

    out = xr.merge(out)
    out.attrs = ds.attrs

    return out


def _get_plotting_positions(data_array, alpha=0.4, beta=0.4, return_period=True):
    """Calculate plotting positions for data.

    Parameters
    ----------
    data_array : xarray DataArray
        Input data.
    alpha : float, optional
        Plotting position parameter, by default 0.4.
    beta : float, optional
        Plotting position parameter, by default 0.4.
    return_period : bool, optional
        If True, return periods instead of probabilities, by default True.

    Returns
    -------
    xarray DataArray
        The data with plotting positions assigned.

    Notes
    -----
    See scipy.stats.mstats.plotting_positions for typical values for alpha and beta. (0.4, 0.4) : approximately quantile unbiased (Cunnane).
    """
    data_copy = data_array.copy(deep=True)

    def vec_plotting_positions(vec_data, alpha=0.4, beta=0.4):
        """Calculate plotting positions for vectorized data.

        Parameters
        ----------
        vec_data : ndarray
            Input data, with time dimension first.
        alpha : float, optional
            Plotting position parameter.
        beta : float, optional
            Plotting position parameter.

        Returns
        -------
        ndarray
            Array with plotting positions assigned to valid data points,
            and NaNs assigned to invalid data points.
        """
        out = []
        if vec_data.ndim == 1:
            valid = ~np.isnan(vec_data)
            pp = plotting_positions(vec_data[valid], alpha, beta)

            out = np.full_like(vec_data.astype(float), np.nan)
            out[valid] = pp

        else:
            for data in vec_data:
                valid = ~np.isnan(data)
                pp = plotting_positions(data[valid], alpha, beta)

                pp_full = np.full_like(data.astype(float), np.nan)
                pp_full[valid] = pp

                out.append(pp_full)
        return out

    pp = xr.apply_ufunc(
        vec_plotting_positions,
        data_array,
        alpha,
        beta,
        input_core_dims=[["time"], [], []],
        output_core_dims=[["time"]],
    )

    if return_period:
        pp = -1 / (pp - 1)

    for name in pp.data_vars:
        pp = pp.rename({name: name + "_pp"})
        pp[name] = data_copy[name]

    return pp


def _prepare_plots(params, xmin=1, xmax=10000, npoints=100, log=True):
    """Prepare x-values for plotting frequency analysis results.

    Parameters
    ----------
    params : xarray DataArray
        Input data.
    xmin : float, optional
        Minimum x value, by default 1.
    xmax : float, optional
        Maximum x value, by default 10000.
    npoints : int, optional
        Number of x values, by default 100.
    log : bool, optional
        If True, return log-spaced x values, by default True.

    Returns
    -------
    xarray DataArray
        The data with plotting positions assigned.
    """
    if log:
        x = np.logspace(np.log10(xmin), np.log10(xmax), npoints, endpoint=True)
    else:
        x = np.linspace(xmin, xmax, npoints)
    return parametric_quantiles(params, x)



# Created on Tue Dec 12 21:55:25 2023
# @author: Richard Arsenault
"""Objective function package for xhydro, for calibration and model evaluation.

This package provides a flexible suite of popular objective function metrics in
hydrological modelling and hydrological model calibration. The main function
'get_objective_function' returns the value of the desired objective function
while allowing users to customize many aspects:

    1-  Select the objective function to run;
    2-  Allow providing a mask to remove certain elements from the objective
    function calculation (e.g. for odd/even year calibration, or calibration
    on high or low flows only, or any custom setup).
    3-  Apply a transformation on the flows to modify the behaviour of the
    objective function calculation (e.g taking the log, inverse or square
    root transform of the flows before computing the objective function).

This function also contains some tools and inputs reserved for the calibration
toolbox, such as the ability to take the negative of the objective function to
maximize instead of minimize a metric according to the needs of the optimizing
algorithm.
"""

# Import packages
import numpy as np
import xarray as xr

__all__ = ["get_objective_function", "transform_flows"]


def get_objective_function(
    qobs: np.array,
    qsim: np.array,
    obj_func: str = "rmse",
    take_negative: bool = False,
    mask: np.array = None,
    transform: str = None,
    epsilon: float = None,
):
    """Entrypoint function for the objective function calculation.

    More can be added by adding the function to this file and adding the option
    in this function.

    NOTE: All data corresponding to NaN values in the observation set are
          removed from the calculation. If a mask is passed, it must be the
          same size as the qsim and qobs vectors. If any NaNs are present in
          the qobs dataset, all corresponding data in the qobs, qsim and mask
          will be removed prior to passing to the processing function.

    Parameters
    ----------
    qobs : array_like
        Vector containing the Observed streamflow to be used in the objective
        function calculation. It is the target to attain.

    qsim : array_like
        Vector containing the Simulated streamflow as generated by the hydro-
        logical model. It is modified by changing parameters and resumulating
        the hydrological model.

    obj_func : str
        String representing the objective function to use in the calibration.
        Options must be one of the accepted objective functions:

            - "abs_bias" : Absolute value of the "bias" metric
            - "abs_pbias": Absolute value of the "pbias" metric
            - "abs_volume_error" : Absolute value of the volume_error metric
            - "agreement_index": Index of agreement
            - "bias" : Bias metric
            - "correlation_coeff": Correlation coefficient
            - "kge" : Kling Gupta Efficiency metric (2009 version)
            - "kge_mod" : Kling Gupta Efficiency metric (2012 version)
            - "mae": Mean Absolute Error metric
            - "mare": Mean Absolute Relative Error metric
            - "mse" : Mean Square Error metric
            - "nse": Nash-Sutcliffe Efficiency metric
            - "pbias" : Percent bias (relative bias)
            - "r2" : r-squared, i.e. square of correlation_coeff.
            - "rmse" : Root Mean Square Error
            - "rrmse" : Relative Root Mean Square Error (RMSE-to-mean ratio)
            - "rsr" : Ratio of RMSE to standard deviation.
            - "volume_error": Total volume error over the period.

        The default is 'rmse'.

    take_negative : bool
        Used to force the objective function to be multiplied by minus one (-1)
        such that it is possible to maximize it if the optimizer is a minimizer
        and vice-versa. Should always be set to False unless required by an
        optimization setup, which is handled internally and transparently to
        the user. The default is False.

    mask : array_like
        Array of 0 or 1 on which the objective function should be applied.
        Values of 1 indicate that the value is included in the calculation, and
        values of 0 indicate that the value is excluded and will have no impact
        on the objective function calculation. This can be useful for specific
        optimization strategies such as odd/even year calibration, seasonal
        calibration or calibration based on high/low flows. The default is None
        and all data are preserved.

    transform : str
        Indicates the type of transformation required. Can be one of the
        following values:

            - "sqrt" : Square root transformation of the flows [sqrt(Q)]
            - "log" : Logarithmic transformation of the flows [log(Q)]
            - "inv" : Inverse transformation of the flows [1/Q]

        The default value is "None", by which no transformation is performed.

    epsilon : float
        Indicates the perturbation to add to the flow time series during a
        transformation to avoid division by zero and logarithmic transformation.
        The perturbation is equal to:

            perturbation = epsilon * mean(qobs)

        The default value is 0.01.

    Returns
    -------
    float
        Value of the selected objective function (obj_fun).
    """
    # List of available objective functions
    obj_func_dict = {
        "abs_bias": _abs_bias,
        "abs_pbias": _abs_pbias,
        "abs_volume_error": _abs_volume_error,
        "agreement_index": _agreement_index,
        "bias": _bias,
        "correlation_coeff": _correlation_coeff,
        "kge": _kge,
        "kge_mod": _kge_mod,
        "mae": _mae,
        "mare": _mare,
        "mse": _mse,
        "nse": _nse,
        "pbias": _pbias,
        "r2": _r2,
        "rmse": _rmse,
        "rrmse": _rrmse,
        "rsr": _rsr,
        "volume_error": _volume_error,
    }

    # If we got a dataset, change to np.array
    # FIXME: Implement a more flexible method
    if isinstance(qsim, xr.Dataset):
        qsim = qsim["qsim"]

    if isinstance(qobs, xr.Dataset):
        qobs = qobs["qobs"]

    # Basic error checking
    if qobs.shape[0] != qsim.shape[0]:
        raise ValueError("Observed and Simulated flows are not of the same size.")

    # Check mask size and contents
    if mask is not None:
        # Size
        if qobs.shape[0] != mask.shape[0]:
            raise ValueError("Mask is not of the same size as the streamflow data.")

        # All zero or one?
        if not np.setdiff1d(np.unique(mask), np.array([0, 1])).size == 0:
            raise ValueError("Mask contains values other than 0 or 1. Please modify.")

    # Check that the objective function is in the list of available methods
    if obj_func not in obj_func_dict:
        raise ValueError(
            "Selected objective function is currently unavailable. "
            + "Consider contributing to our project at: "
            + "github.com/hydrologie/xhydro"
        )

    # Ensure there are no NaNs in the dataset
    if mask is not None:
        mask = mask[~np.isnan(qobs)]
    qsim = qsim[~np.isnan(qobs)]
    qobs = qobs[~np.isnan(qobs)]

    # Apply mask before trasform
    if mask is not None:
        qsim = qsim[mask == 1]
        qobs = qobs[mask == 1]
        mask = mask[mask == 1]

    # Transform data if needed
    if transform is not None:
        qsim, qobs = transform_flows(qsim, qobs, transform, epsilon)

    # Compute objective function by switching to the correct algorithm. Ensure
    # that the function name is the same as the obj_func tag or this will fail.
    function_call = obj_func_dict[obj_func]
    obj_fun_val = function_call(qsim, qobs)

    # Take the negative value of the Objective Function to return to the
    # optimizer.
    if take_negative:
        obj_fun_val = obj_fun_val * -1

    return obj_fun_val


def _get_objfun_minimize_or_maximize(obj_func: str):
    """Check whether the objective function needs to be maximized or minimized.

    Returns a boolean value, where True means it should be maximized, and False
    means that it should be minimized. Objective functions other than those
    programmed here will raise an error.

    Parameters
    ----------
    obj_func : str
        Label of the desired objective function.

    Returns
    -------
    bool
        Indicator of if the objective function needs to be maximized.
    """
    # Define metrics that need to be maximized:
    if obj_func in [
        "agreement_index",
        "correlation_coeff",
        "kge",
        "kge_mod",
        "nse",
        "r2",
    ]:
        maximize = True

    # Define the metrics that need to be minimized:
    elif obj_func in [
        "abs_bias",
        "abs_pbias",
        "abs_volume_error",
        "mae",
        "mare",
        "mse",
        "rmse",
        "rrmse",
        "rsr",
    ]:
        maximize = False

    # Check for the metrics that exist but cannot be used for optimization
    elif obj_func in ["bias", "pbias", "volume_error"]:
        raise ValueError(
            "The bias, pbias and volume_error metrics cannot be minimized or maximized. \
                 Please use the abs_bias, abs_pbias and abs_volume_error instead."
        )
    else:
        raise NotImplementedError("The objective function is unknown.")

    return maximize


def _get_optimizer_minimize_or_maximize(algorithm: str):
    """Find the direction in which the optimizer searches.

    Some optimizers try to maximize the objective function value, and others
    try to minimize it. Since our objective functions include some that need to
    be maximized and others minimized, it is imperative to ensure that the
    optimizer/objective-function pair work in tandem.

    Parameters
    ----------
    algorithm : str
        Name of the optimizing algorithm.

    Returns
    -------
    bool
        Indicator of the direction of the optimizer search, True to maximize.
    """
    # Define metrics that need to be maximized:
    if algorithm in [
        "DDS",
    ]:
        maximize = True

    # Define the metrics that need to be minimized:
    elif algorithm in [
        "SCEUA",
    ]:
        maximize = False

    # Any other optimizer at this date
    else:
        raise NotImplementedError("The optimization algorithm is unknown.")

    return maximize


def transform_flows(
    qsim: np.array, qobs: np.array, transform: str = None, epsilon: float = 0.01
):
    """
    Transform flows before computing the objective function.

    It is used to transform flows such that the objective function is computed
    on a transformed flow metric rather than on the original units of flow
    (ex: inverse, log-transformed, square-root)

    Parameters
    ----------
    qsim : array_like
        Simulated streamflow vector.
    qobs : array_like
        Observed streamflow vector.

    transform : str
        Indicates the type of transformation required. Can be one of the
        following values:

            - "sqrt" : Square root transformation of the flows [sqrt(Q)]
            - "log" : Logarithmic transformation of the flows [log(Q)]
            - "inv" : Inverse transformation of the flows [1/Q]

        The default value is "None", by which no transformation is performed.

    epsilon : float
        Indicates the perturbation to add to the flow time series during a
        transformation to avoid division by zero and logarithmic transformation.
        The perturbation is equal to:

            perturbation = epsilon * mean(qobs)

        The default value is 0.01.

    Returns
    -------
    qsim : array_like
        Transformed simulated flow according to user request.
    qobs : array_like
        Transformed observed flow according to user request.
    """
    # Quick check
    if transform is not None:
        if transform not in ["log", "inv", "sqrt"]:
            raise NotImplementedError("Flow transformation method not recognized.")

    # Transform the flow series if required
    if transform == "log":  # log transformation
        epsilon = epsilon * np.nanmean(qobs)
        qobs, qsim = np.log(qobs + epsilon), np.log(qsim + epsilon)

    elif transform == "inv":  # inverse transformation
        epsilon = epsilon * np.nanmean(qobs)
        qobs, qsim = 1.0 / (qobs + epsilon), 1.0 / (qsim + epsilon)

    elif transform == "sqrt":  # square root transformation
        qobs, qsim = np.sqrt(qobs), np.sqrt(qsim)

    # Return the flows after transformation (or original if no transform)
    return qsim, qobs


"""
BEGIN OBJECTIVE FUNCTIONS DEFINITIONS
"""


def _abs_bias(qsim: np.array, qobs: np.array):
    """
    Absolute bias metric.

    Parameters
    ----------
    qsim : array_like
        Simulated streamflow vector.
    qobs : array_like
        Observed streamflow vector.

    Returns
    -------
    float
        Absolute value of the "bias" metric. This metric is useful when calibrating
        on the bias, because bias should aim to be 0 but can take large positive
        or negative values. Taking the absolute value of the bias will let the
        optimizer minimize the value to zero.

    Notes
    -----
    The abs_bias should be MINIMIZED.
    """
    return np.abs(_bias(qsim, qobs))


def _abs_pbias(qsim: np.array, qobs: np.array):
    """
    Absolute pbias metric.

    Parameters
    ----------
    qsim : array_like
        Simulated streamflow vector.
    qobs : array_like
        Observed streamflow vector.

    Returns
    -------
    float
        The absolute value of the "pbias" metric. This metric is useful when
        calibrating on the pbias, because pbias should aim to be 0 but can take
        large positive or negative values. Taking the absolute value of the
        pbias will let the optimizer minimize the value to zero.

    Notes
    -----
    The abs_pbias should be MINIMIZED.
    """
    return np.abs(_pbias(qsim, qobs))


def _abs_volume_error(qsim: np.array, qobs: np.array):
    """
    Absolute value of the volume error metric.

    Parameters
    ----------
    qsim : array_like
        Simulated streamflow vector.
    qobs : array_like
        Observed streamflow vector.

    Returns
    -------
    float
        The absolute value of the "volume_error" metric. This metric is useful
        when calibrating on the volume_error, because volume_error should aim
        to be 0 but can take large positive or negative values. Taking the
        absolute value of the volume_error will let the optimizer minimize the
        value to zero.

    Notes
    -----
    The abs_volume_error should be MINIMIZED.
    """
    return np.abs(_volume_error(qsim, qobs))


def _agreement_index(qsim: np.array, qobs: np.array):
    """
    Index of agreement metric.

    Parameters
    ----------
    qsim : array_like
        Simulated streamflow vector.
    qobs : array_like
        Observed streamflow vector.

    Returns
    -------
    float
        The agreement index of Willmott (1981). Varies between 0 and 1.

    Notes
    -----
    The Agreement index should be MAXIMIZED.
    """
    # Decompose into clearer chunks
    a = np.sum((qobs - qsim) ** 2)
    b = np.abs(qsim - np.mean(qobs)) + np.abs(qobs - np.mean(qobs))
    c = np.sum(b**2)

    return 1 - (a / c)


def _bias(qsim: np.array, qobs: np.array):
    """
    The bias metric.

    Parameters
    ----------
    qsim : array_like
        Simulated streamflow vector.
    qobs : array_like
        Observed streamflow vector.

    Returns
    -------
    float
        The bias in the simulation. Can be negative or positive and gives the
        average error between the observed and simulated flows. This
        interpretation uses the definition that a positive bias value means
        that the simulation overestimates the true value (as opposed to many
        other sources on bias calculations that use the contrary interpretation).

    Notes
    -----
    BIAS SHOULD AIM TO BE ZERO AND SHOULD NOT BE USED FOR CALIBRATION. FOR
    CALIBRATION, USE "abs_bias" TO TAKE THE ABSOLUTE VALUE.
    """
    return np.mean(qsim - qobs)


def _correlation_coeff(qsim: np.array, qobs: np.array):
    """
    Correlation coefficient metric.

    Parameters
    ----------
    qsim : array_like
        Simulated streamflow vector.
    qobs : array_like
        Observed streamflow vector.

    Returns
    -------
    float
        The correlation coefficient.

    Notes
    -----
    The correlation_coeff should be MAXIMIZED.
    """
    return np.corrcoef(qobs, qsim)[0, 1]


def _kge(qsim: np.array, qobs: np.array):
    """
    Kling-Gupta efficiency metric (2009 version).

    Parameters
    ----------
    qsim : array_like
        Simulated streamflow vector.
    qobs : array_like
        Observed streamflow vector.

    Returns
    -------
    float
        The Kling-Gupta Efficiency (KGE) metric of 2009. It can take values
        from -inf to 1 (best case).

    Notes
    -----
    The KGE should be MAXIMIZED.
    """
    # This pops up a lot, precalculate.
    Qsim_mean = np.mean(qsim)
    qobs_mean = np.mean(qobs)

    # Calculate the components of KGE
    r_num = np.sum((qsim - Qsim_mean) * (qobs - qobs_mean))
    r_den = np.sqrt(np.sum((qsim - Qsim_mean) ** 2) * np.sum((qobs - qobs_mean) ** 2))
    r = r_num / r_den
    a = np.std(qsim) / np.std(qobs)
    b = np.sum(qsim) / np.sum(qobs)

    # Calculate the KGE
    kge = 1 - np.sqrt((r - 1) ** 2 + (a - 1) ** 2 + (b - 1) ** 2)

    return kge


def _kge_mod(qsim: np.array, qobs: np.array):
    """
    Kling-Gupta efficiency metric (2012 version).

    Parameters
    ----------
    qsim : array_like
        Simulated streamflow vector.
    qobs : array_like
        Observed streamflow vector.

    Returns
    -------
    float
        The modified Kling-Gupta Efficiency (KGE) metric of 2012. It can take
        values from -inf to 1 (best case).

    Notes
    -----
    The kge_mod should be MAXIMIZED.
    """
    # These pop up a lot, precalculate
    Qsim_mean = np.mean(qsim)
    qobs_mean = np.mean(qobs)

    # Calc KGE components
    r_num = np.sum((qsim - Qsim_mean) * (qobs - qobs_mean))
    r_den = np.sqrt(np.sum((qsim - Qsim_mean) ** 2) * np.sum((qobs - qobs_mean) ** 2))
    r = r_num / r_den
    g = (np.std(qsim) / Qsim_mean) / (np.std(qobs) / qobs_mean)
    b = np.mean(qsim) / np.mean(qobs)

    # Calc the modified KGE metric
    kge_mod = 1 - np.sqrt((r - 1) ** 2 + (g - 1) ** 2 + (b - 1) ** 2)

    return kge_mod


def _mae(qsim: np.array, qobs: np.array):
    """
    Mean absolute error metric.

    Parameters
    ----------
    qsim : array_like
        Simulated streamflow vector.
    qobs : array_like
        Observed streamflow vector.

    Returns
    -------
    float
        Mean Absolute Error. It can be interpreted as the average error
        (absolute) between observations and simulations for any time step.

    Notes
    -----
    The mae should be MINIMIZED.
    """
    return np.mean(np.abs(qsim - qobs))


def _mare(qsim: np.array, qobs: np.array):
    """
    Mean absolute relative error metric.

    Parameters
    ----------
    qsim : array_like
        Simulated streamflow vector.
    qobs : array_like
        Observed streamflow vector.

    Returns
    -------
    float
        Mean Absolute Relative Error. For streamflow, where qobs is always zero
        or positive, the MARE is always positive.

    Notes
    -----
    The mare should be MINIMIZED.
    """
    return np.sum(np.abs(qobs - qsim)) / np.sum(qobs)


def _mse(qsim: np.array, qobs: np.array):
    """
    Mean square error metric.

    Parameters
    ----------
    qsim : array_like
        Simulated streamflow vector.
    qobs : array_like
        Observed streamflow vector.

    Returns
    -------
    float
        Mean Square Error. It is the sum of squared errors for each day divided
        by the total number of days. Units are thus squared units, and the best
        possible value is 0.

    Notes
    -----
    The mse should be MINIMIZED.
    """
    return np.mean((qobs - qsim) ** 2)


def _nse(qsim: np.array, qobs: np.array):
    """
    Nash-Sutcliffe efficiency metric.

    Parameters
    ----------
    qsim : array_like
        Simulated streamflow vector.
    qobs : array_like
        Observed streamflow vector.

    Returns
    -------
    float
        Nash-Sutcliffe Efficiency (NSE) metric. It can take values from -inf to
        1, with 0 being as good as using the mean observed flow as the estimator.

    Notes
    -----
    The nse should be MAXIMIZED.
    """
    num = np.sum((qobs - qsim) ** 2)
    den = np.sum((qobs - np.mean(qobs)) ** 2)

    return 1 - (num / den)


def _pbias(qsim: np.array, qobs: np.array):
    """
    Percent bias metric.

    Parameters
    ----------
    qsim : array_like
        Simulated streamflow vector.
    qobs : array_like
        Observed streamflow vector.

    Returns
    -------
    float
        Percent bias. Can be negative or positive and gives the average
        relative error between the observed and simulated flows. This
        interpretation uses the definition that a positive bias value means
        that the simulation overestimates the true value (as opposed to many
        other sources on bias calculations that use the contrary interpretation).

    Notes
    -----
    PBIAS SHOULD AIM TO BE ZERO AND SHOULD NOT BE USED FOR CALIBRATION. FOR
    CALIBRATION, USE "abs_pbias" TO TAKE THE ABSOLUTE VALUE.
    """
    return (np.sum(qsim - qobs) / np.sum(qobs)) * 100


def _r2(qsim: np.array, qobs: np.array):
    """
    The r-squred metric.

    Parameters
    ----------
    qsim : array_like
        Simulated streamflow vector.
    qobs : array_like
        Observed streamflow vector.

    Returns
    -------
    float
        The r-squared (R2) metric equal to the square of the correlation
        coefficient.

    Notes
    -----
    The r2 should be MAXIMIZED.
    """
    return _correlation_coeff(qsim, qobs) ** 2


def _rmse(qsim: np.array, qobs: np.array):
    """
    Root mean square error metric.

    Parameters
    ----------
    qsim : array_like
        Simulated streamflow vector.
    qobs : array_like
        Observed streamflow vector.

    Returns
    -------
    float
        Root Mean Square Error. Units are the same as the timeseries data
        (ex. m3/s). It can take zero or positive values.

    Notes
    -----
    The rmse should be MINIMIZED.
    """
    return np.sqrt(np.mean((qobs - qsim) ** 2))


def _rrmse(qsim: np.array, qobs: np.array):
    """
    Relative root mean square error (ratio of rmse to mean) metric.

    Parameters
    ----------
    qsim : array_like
        Simulated streamflow vector.
    qobs : array_like
        Observed streamflow vector.

    Returns
    -------
    float
        Ratio of the RMSE to the mean of the observations. It allows scaling
        RMSE values to compare results between time series of different
        magnitudes (ex. flows from small and large watersheds). Also known as
        the CVRMSE.

    Notes
    -----
    The rrmse should be MINIMIZED.
    """
    return _rmse(qsim, qobs) / np.mean(qobs)


def _rsr(qsim: np.array, qobs: np.array):
    """
    Ratio of root mean square error to standard deviation metric.

    Parameters
    ----------
    qsim : array_like
        Simulated streamflow vector.
    qobs : array_like
        Observed streamflow vector.

    Returns
    -------
    float
        Root Mean Square Error (RMSE) divided by the standard deviation of the
        observations. Also known as the "Ratio of the Root Mean Square Error to
        the Standard Deviation of Observations".

    Notes
    -----
    The rsr should be MINIMIZED.
    """
    return _rmse(qobs, qsim) / np.std(qobs)


def _volume_error(qsim: np.array, qobs: np.array):
    """
    Volume error metric.

    Parameters
    ----------
    qsim : array_like
        Simulated streamflow vector.
    qobs : array_like
        Observed streamflow vector.

    Returns
    -------
    float
        Total error in terms of volume over the entire period. Expressed in
        terms of the same units as input data, so for flow rates it is
        important to multiply by the duration of the time-step to obtain actual
        volumes.

    Notes
    -----
    The volume_error should be MINIMIZED.
    """
    return np.sum(qsim - qobs) / np.sum(qobs)


"""
ADD OBJECTIVE FUNCTIONS HERE
"""



import numpy as np
import pytest
from xclim.testing.helpers import test_timeseries as timeseries

import xhydro as xh


# Smoke test for xscen functions that are imported into xhydro
def test_xscen_imported():
    assert callable(getattr(xh.indicators, "compute_indicators"))


class TestComputeVolume:
    @pytest.mark.parametrize("freq", ["D", "YS"])
    def test_compute_volume(self, freq):
        tile = 365 if freq == "D" else 1
        da = timeseries(
            np.tile(np.arange(1, tile + 1), 3),
            variable="streamflow",
            start="2001-01-01",
            freq=freq,
        )

        out = xh.indicators.compute_volume(da, attrs={"long_name": "Foo"})
        mult = 86400 if freq == "D" else 86400 * 365
        np.testing.assert_array_equal(out, da * mult)
        assert out.attrs["long_name"] == "Foo"
        assert out.attrs["cell_methods"] == "time: sum"
        assert out.attrs["description"] == "Volume of water"
        assert out.attrs["units"] == "m3"

    def test_units(self):
        da = timeseries(
            np.tile(np.arange(1, 366), 3),
            variable="streamflow",
            start="2001-01-01",
            freq="D",
        )

        out_m3 = xh.indicators.compute_volume(da)
        out_hm3 = xh.indicators.compute_volume(da, out_units="hm3")

        assert out_m3.attrs["units"] == "m3"
        assert out_hm3.attrs["units"] == "hm3"

        np.testing.assert_array_equal(out_m3 * 1e-6, out_hm3)


class TestGetYearlyOp:
    ds = timeseries(
        np.arange(1, 365 * 3 + 1),
        variable="streamflow",
        start="2001-01-01",
        freq="D",
        as_dataset=True,
    )

    @pytest.mark.parametrize("op", ["max", "min"])
    def test_get_yearly_op(self, op):
        timeargs = {
            "annual": {},
            "winterdate": {"date_bounds": ["12-01", "02-28"], "freq": "YS-DEC"},
            "winterdoy": {"doy_bounds": [335, 59], "freq": "YS-DEC"},
            "winterdjf": {"season": ["DJF"], "freq": "YS-DEC"},
            "summer": {"doy_bounds": [200, 300]},
        }

        out = xh.indicators.get_yearly_op(self.ds, op=op, timeargs=timeargs)
        assert all(["streamflow" in v for v in out.data_vars])
        assert len(out.data_vars) == len(timeargs)

        if op == "max":
            np.testing.assert_array_equal(
                out.streamflow_max_annual,
                np.add(np.tile(365, 3), np.array([0, 365, 365 * 2])),
            )
            np.testing.assert_array_equal(
                out.streamflow_max_summer,
                np.add(np.tile(300, 3), np.array([0, 365, 365 * 2])),
            )
            np.testing.assert_array_equal(
                out.streamflow_max_winterdate,
                np.add(
                    np.array([365 + 59, 365 + 59, 365]), np.array([0, 365, 365 * 2])
                ),
            )
            np.testing.assert_array_equal(
                out.streamflow_max_winterdoy, out.streamflow_max_winterdate
            )
            np.testing.assert_array_equal(
                out.streamflow_max_winterdjf, out.streamflow_max_winterdate
            )
        elif op == "min":
            np.testing.assert_array_equal(
                out.streamflow_min_annual,
                np.add(np.tile(1, 3), np.array([0, 365, 365 * 2])),
            )
            np.testing.assert_array_equal(
                out.streamflow_min_summer,
                np.add(np.tile(200, 3), np.array([0, 365, 365 * 2])),
            )
            np.testing.assert_array_equal(
                out.streamflow_min_winterdate,
                np.add(np.tile(335, 3), np.array([0, 365, 365 * 2])),
            )
            np.testing.assert_array_equal(
                out.streamflow_min_winterdoy, out.streamflow_min_winterdate
            )
            np.testing.assert_array_equal(
                out.streamflow_min_winterdjf, out.streamflow_min_winterdate
            )

    def test_missing(self):
        timeargs = {"winterdate": {"date_bounds": ["12-01", "02-28"], "freq": "YS-DEC"}}
        out = xh.indicators.get_yearly_op(
            self.ds,
            op="max",
            timeargs=timeargs,
            missing="pct",
            missing_options={"tolerance": 0.1},
        )

        np.testing.assert_array_equal(
            out.streamflow_max_winterdate,
            np.add(np.array([365 + 59, 365 + 59, np.nan]), np.array([0, 365, 365 * 2])),
        )

    def test_window(self):
        out = xh.indicators.get_yearly_op(self.ds, op="max", window=2)

        assert all(["streamflow2" in v for v in out.data_vars])
        np.testing.assert_array_equal(
            out.streamflow2_max_annual, np.array([364.5, 729.5, 1094.5])
        )

    def test_sum(self):
        ds = timeseries(
            np.arange(1, 365 * 3 + 1),
            variable="streamflow",
            start="2001-01-01",
            freq="D",
            as_dataset=True,
        )
        ds["volume"] = xh.indicators.compute_volume(ds.streamflow)
        ds["volume"] = ds["volume"].where(
            ~((ds.time.dt.month == 1) & (ds.time.dt.day == 3))
        )

        timeargs = {
            "annual": {},
            "winterdate": {"date_bounds": ["12-01", "02-28"], "freq": "YS-DEC"},
            "summer": {"doy_bounds": [200, 300]},
        }
        out_sum = xh.indicators.get_yearly_op(
            ds, input_var="volume", op="sum", timeargs=timeargs
        )
        out_interp = xh.indicators.get_yearly_op(
            ds, input_var="volume", op="sum", timeargs=timeargs, interpolate_na=True
        )

        ans = {
            "annual": np.array(
                [
                    np.sum(np.arange(1, 365 + 1)) * 86400,
                    np.sum(np.arange(1 + 365, 365 + 365 + 1)) * 86400,
                    np.sum(np.arange(1 + 730, 365 + 730 + 1)) * 86400,
                ]
            ),
            "summer": np.array(
                [
                    np.sum(np.arange(200, 300 + 1)) * 86400,
                    np.sum(np.arange(200 + 365, 300 + 365 + 1)) * 86400,
                    np.sum(np.arange(200 + 730, 300 + 730 + 1)) * 86400,
                ]
            ),
            "winterdate": np.array(
                [
                    np.sum(
                        np.concatenate(
                            (np.arange(335, 365 + 1), np.arange(1 + 365, 59 + 365 + 1))
                        )
                    )
                    * 86400,
                    np.sum(
                        np.concatenate(
                            (
                                np.arange(335 + 365, 365 + 365 + 1),
                                np.arange(1 + 730, 59 + 730 + 1),
                            )
                        )
                    )
                    * 86400,
                    np.sum(np.arange(335 + 730, 365 + 730 + 1)) * 86400,
                ]
            ),
        }

        assert all(["volume" in v for v in out_interp.data_vars])
        np.testing.assert_array_equal(
            out_interp.volume_sum_summer, out_sum.volume_sum_summer
        )
        np.testing.assert_array_equal(out_interp.volume_sum_annual, ans["annual"])
        np.testing.assert_array_equal(out_interp.volume_sum_summer, ans["summer"])
        np.testing.assert_array_equal(
            out_interp.volume_sum_winterdate, ans["winterdate"]
        )

        np.testing.assert_array_equal(
            out_sum.volume_sum_annual, (ans["annual"] - np.array([3, 368, 733]) * 86400)
        )
        np.testing.assert_array_equal(out_sum.volume_sum_summer, ans["summer"])
        np.testing.assert_array_equal(
            out_sum.volume_sum_winterdate,
            ans["winterdate"] - np.array([368, 733, 0]) * 86400,
        )

    def test_errors(self):
        with pytest.raises(ValueError, match="Operation foo is not supported."):
            xh.indicators.get_yearly_op(self.ds, op="foo")
        with pytest.raises(ValueError, match="Cannot use a rolling window"):
            xh.indicators.get_yearly_op(self.ds, op="sum", window=2)
        with pytest.raises(ValueError, match="Frequency D is not supported"):
            xh.indicators.get_yearly_op(
                self.ds, op="max", timeargs={"annual": {"freq": "D"}}
            )
        with pytest.raises(ValueError, match="Only one indexer"):
            xh.indicators.get_yearly_op(
                self.ds,
                op="max",
                timeargs={"annual": {"season": ["DJF"], "doy_bounds": [200, 300]}},
            )
        with pytest.warns(UserWarning, match="The frequency is not YS-DEC"):
            xh.indicators.get_yearly_op(
                self.ds, op="max", timeargs={"annual": {"season": ["DJF"]}}
            )
        with pytest.warns(UserWarning, match="The bounds wrap around the year"):
            xh.indicators.get_yearly_op(
                self.ds,
                op="max",
                timeargs={"annual": {"date_bounds": ["06-15", "06-14"]}},
            )
        with pytest.warns(UserWarning, match="but the bounds"):
            xh.indicators.get_yearly_op(
                self.ds,
                op="max",
                timeargs={
                    "annual": {"date_bounds": ["06-01", "04-30"], "freq": "YS-DEC"}
                },
            )



from pathlib import Path

import pytest

import xhydro.testing.utils as xhu


@pytest.mark.requires_docs
def test_publish_release_notes(tmp_path):
    temp_md_filename = tmp_path.joinpath("version_info.md")
    xhu.publish_release_notes(
        style="md",
        file=temp_md_filename,
        changes=Path(__file__).parent.parent.joinpath("CHANGES.rst"),
    )

    with open(temp_md_filename) as f:
        changelog = f.read()

    assert changelog.startswith("# Changelog")
    assert ":user:`" not in changelog
    assert ":issue:`" not in changelog
    assert ":pull:`" not in changelog

    temp_rst_filename = tmp_path.joinpath("version_info.rst")
    xhu.publish_release_notes(
        style="rst",
        file=temp_rst_filename,
        changes=Path(__file__).parent.parent.joinpath("CHANGES.rst"),
    )
    with open(temp_rst_filename) as f:
        changelog_rst = f.read()
    assert changelog_rst.startswith("=========\nChangelog\n=========")



"""Test suite for the calibration algorithm in calibration.py."""

# Also tests the dummy model implementation.
import numpy as np
import pytest

from xhydro.modelling.calibration import perform_calibration
from xhydro.modelling.hydrological_modelling import _dummy_model
from xhydro.modelling.obj_funcs import get_objective_function, transform_flows


def test_spotpy_calibration():
    """Make sure the calibration works under possible test cases."""
    bounds_low = np.array([0, 0, 0])
    bounds_high = np.array([10, 10, 10])

    model_config = {
        "precip": np.array([10, 11, 12, 13, 14, 15]),
        "temperature": np.array([10, 3, -5, 1, 15, 0]),
        "qobs": np.array([120, 130, 140, 150, 160, 170]),
        "drainage_area": np.array([10]),
        "model_name": "Dummy",
    }

    mask = np.array([0, 0, 0, 0, 1, 1])

    best_parameters, best_simulation, best_objfun = perform_calibration(
        model_config,
        "mae",
        bounds_low=bounds_low,
        bounds_high=bounds_high,
        evaluations=1000,
        algorithm="DDS",
        mask=mask,
        sampler_kwargs=dict(trials=1),
    )

    # Test that the results have the same size as expected (number of parameters)
    assert len(best_parameters) == len(bounds_high)

    # Test that the objective function is calculated correctly
    objfun = get_objective_function(
        model_config["qobs"],
        best_simulation,
        obj_func="mae",
        mask=mask,
    )

    assert objfun == best_objfun

    # Test dummy model response
    model_config["parameters"] = [5, 5, 5]
    qsim = _dummy_model(model_config)
    assert qsim["qsim"].values[3] == 3500.00

    # Also test to ensure SCEUA and take_minimize is required.
    best_parameters_sceua, best_simulation, best_objfun = perform_calibration(
        model_config,
        "mae",
        bounds_low=bounds_low,
        bounds_high=bounds_high,
        evaluations=10,
        algorithm="SCEUA",
    )

    assert len(best_parameters_sceua) == len(bounds_high)

    # Also test to ensure SCEUA and take_minimize is required.
    best_parameters_negative, best_simulation, best_objfun = perform_calibration(
        model_config,
        "nse",
        bounds_low=bounds_low,
        bounds_high=bounds_high,
        evaluations=10,
        algorithm="SCEUA",
    )
    assert len(best_parameters_negative) == len(bounds_high)

    # Test to see if transform works
    best_parameters_transform, best_simulation, best_objfun = perform_calibration(
        model_config,
        "nse",
        bounds_low=bounds_low,
        bounds_high=bounds_high,
        evaluations=10,
        algorithm="SCEUA",
        transform="inv",
        epsilon=0.01,
    )
    assert len(best_parameters_transform) == len(bounds_high)


def test_calibration_failure_mode_unknown_optimizer():
    """Test for maximize-minimize failure mode:
    use "OTHER" optimizer, i.e. an unknown optimizer. Should fail.
    """
    bounds_low = np.array([0, 0, 0])
    bounds_high = np.array([10, 10, 10])
    model_config = {
        "precip": np.array([10, 11, 12, 13, 14, 15]),
        "temperature": np.array([10, 3, -5, 1, 15, 0]),
        "qobs": np.array([120, 130, 140, 150, 160, 170]),
        "drainage_area": np.array([10]),
        "model_name": "Dummy",
    }
    with pytest.raises(NotImplementedError) as pytest_wrapped_e:
        best_parameters_transform, best_simulation, best_objfun = perform_calibration(
            model_config,
            "nse",
            bounds_low=bounds_low,
            bounds_high=bounds_high,
            evaluations=10,
            algorithm="OTHER",
        )
        assert pytest_wrapped_e.type == NotImplementedError


def test_transform():
    """Test the flow transformer"""
    qsim = np.array([10, 10, 10])
    qobs = np.array([5, 5, 5])

    qsim_r, qobs_r = transform_flows(qsim, qobs, transform="inv", epsilon=0.01)
    np.testing.assert_array_almost_equal(qsim_r[1], 0.0995024, 6)
    np.testing.assert_array_almost_equal(qobs_r[1], 0.1980198, 6)

    qsim_r, qobs_r = transform_flows(qsim, qobs, transform="sqrt")
    np.testing.assert_array_almost_equal(qsim_r[1], 3.1622776, 6)
    np.testing.assert_array_almost_equal(qobs_r[1], 2.2360679, 6)

    qsim_r, qobs_r = transform_flows(qsim, qobs, transform="log", epsilon=0.01)
    np.testing.assert_array_almost_equal(qsim_r[1], 2.3075726, 6)
    np.testing.assert_array_almost_equal(qobs_r[1], 1.6193882, 6)

    # Test Qobs different length than Qsim
    with pytest.raises(NotImplementedError) as pytest_wrapped_e:
        qobs_r, qobs_r = transform_flows(qsim, qobs, transform="a", epsilon=0.01)
        assert pytest_wrapped_e.type == NotImplementedError



#!/usr/bin/env python

"""Tests for `xhydro` package."""

import pathlib
import pkgutil

import pytest


@pytest.fixture
def response():
    """Sample pytest fixture.

    See more at: https://doc.pytest.org/en/latest/explanation/fixtures.html
    """
    # import requests
    # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')
    pass


def test_content(response):
    """Sample pytest test function with the pytest fixture as an argument."""
    # from bs4 import BeautifulSoup
    # assert 'GitHub' in BeautifulSoup(response.content).title.string
    pass


def test_package_metadata():
    """Test the package metadata."""
    project = pkgutil.get_loader("xhydro").get_filename()

    metadata = pathlib.Path(project).resolve().parent.joinpath("__init__.py")

    with open(metadata) as f:
        contents = f.read()
        assert """Thomas-Charles Fortier Filion""" in contents
        assert '__email__ = "tcff_hydro@outlook.com"' in contents
        assert '__version__ = "0.3.4"' in contents



"""Test suite for the objective functions in obj_funcs.py."""

import numpy as np
import pytest

from xhydro.modelling.obj_funcs import (
    _get_objfun_minimize_or_maximize,
    get_objective_function,
)


def test_obj_funcs():
    """Series of tests to test all objective functions with fast test data"""
    qobs = np.array([120, 130, 140, 150, 160, 170])
    qsim = np.array([120, 125, 145, 140, 140, 180])

    # Test that the objective function is calculated correctly
    objfun = get_objective_function(qobs, qsim, obj_func="abs_bias")
    np.testing.assert_array_almost_equal(objfun, 3.3333333333333335, 8)

    objfun = get_objective_function(qobs, qsim, obj_func="abs_pbias")
    np.testing.assert_array_almost_equal(objfun, 2.2988505747126435, 8)

    objfun = get_objective_function(qobs, qsim, obj_func="abs_volume_error")
    np.testing.assert_array_almost_equal(objfun, 0.022988505747126436, 8)

    objfun = get_objective_function(qobs, qsim, obj_func="agreement_index")
    np.testing.assert_array_almost_equal(objfun, 0.9171974522292994, 8)

    objfun = get_objective_function(qobs, qsim, obj_func="bias")
    np.testing.assert_array_almost_equal(objfun, -3.3333333333333335, 8)

    objfun = get_objective_function(qobs, qsim, obj_func="correlation_coeff")
    np.testing.assert_array_almost_equal(objfun, 0.8599102447336393, 8)

    objfun = get_objective_function(qobs, qsim, obj_func="kge")
    np.testing.assert_array_almost_equal(objfun, 0.8077187696552522, 8)

    objfun = get_objective_function(qobs, qsim, obj_func="kge_mod")
    np.testing.assert_array_almost_equal(objfun, 0.7888769531580001, 8)

    objfun = get_objective_function(qobs, qsim, obj_func="mae")
    np.testing.assert_array_almost_equal(objfun, 8.333333333333334, 8)

    objfun = get_objective_function(qobs, qsim, obj_func="mare")
    np.testing.assert_array_almost_equal(objfun, 0.05747126436781609, 8)

    objfun = get_objective_function(qobs, qsim, obj_func="mse")
    np.testing.assert_array_almost_equal(objfun, 108.33333333333333, 8)

    objfun = get_objective_function(qobs, qsim, obj_func="nse")
    np.testing.assert_array_almost_equal(objfun, 0.6285714285714286, 8)

    objfun = get_objective_function(qobs, qsim, obj_func="pbias")
    np.testing.assert_array_almost_equal(objfun, -2.2988505747126435, 8)

    objfun = get_objective_function(qobs, qsim, obj_func="r2")
    np.testing.assert_array_almost_equal(objfun, 0.7394456289978675, 8)

    objfun = get_objective_function(qobs, qsim, obj_func="rmse")
    np.testing.assert_array_almost_equal(objfun, 10.408329997330663, 8)

    objfun = get_objective_function(qobs, qsim, obj_func="rrmse")
    np.testing.assert_array_almost_equal(objfun, 0.07178158618848733, 8)

    objfun = get_objective_function(qobs, qsim, obj_func="rsr")
    np.testing.assert_array_almost_equal(objfun, 0.6094494002200439, 8)

    objfun = get_objective_function(qobs, qsim, obj_func="volume_error")
    np.testing.assert_array_almost_equal(objfun, -0.022988505747126436, 8)


def test_objective_function_failure_data_length():
    """Test for the objective function calculation failure mode:
    qobs and qsim length are different
    """
    with pytest.raises(ValueError) as pytest_wrapped_e:
        _ = get_objective_function(
            np.array([100, 110]),
            np.array([100, 110, 120]),
            obj_func="mae",
        )
        assert pytest_wrapped_e.type == ValueError


def test_objective_function_failure_mask_length():
    """Test for the objective function calculation failure mode:
    qobs and mask length are different
    """
    with pytest.raises(ValueError) as pytest_wrapped_e:
        _ = get_objective_function(
            np.array([100, 100, 100]),
            np.array([100, 110, 120]),
            obj_func="mae",
            mask=np.array([0, 1, 0, 0]),
        )
        assert pytest_wrapped_e.type == ValueError


def test_objective_function_failure_unknown_objfun():
    """Test for the objective function calculation failure mode:
    Objective function is unknown
    """
    with pytest.raises(ValueError) as pytest_wrapped_e:
        _ = get_objective_function(
            np.array([100, 100, 100]),
            np.array([100, 110, 120]),
            obj_func="fake",
        )
        assert pytest_wrapped_e.type == ValueError


def test_objective_function_failure_mask_contents():
    """Test for the objective function calculation failure mode:
    Mask contains other than 0 and 1
    """
    with pytest.raises(ValueError) as pytest_wrapped_e:
        _ = get_objective_function(
            np.array([100, 100, 100]),
            np.array([100, 110, 120]),
            obj_func="mae",
            mask=np.array([0, 0.5, 1]),
        )
        assert pytest_wrapped_e.type == ValueError


def test_maximizer_objfun_failure_modes_bias():
    """Test for maximize-minimize failure mode:
    Use of bias objfun which is unbounded
    """
    with pytest.raises(ValueError) as pytest_wrapped_e:
        _ = _get_objfun_minimize_or_maximize(obj_func="bias")
        assert pytest_wrapped_e.type == ValueError


def test_maximizer_objfun_failure_modes_pbias():
    """Test for maximize-minimize failure mode:
    Use of pbias objfun which is unbounded
    """
    with pytest.raises(ValueError) as pytest_wrapped_e:
        _ = _get_objfun_minimize_or_maximize(obj_func="pbias")
        assert pytest_wrapped_e.type == ValueError


def test_maximizer_objfun_failure_modes_volume_error():
    """Test for maximize-minimize failure mode:
    Use of volume_error objfun which is unbounded
    """
    with pytest.raises(ValueError) as pytest_wrapped_e:
        _ = _get_objfun_minimize_or_maximize(obj_func="volume_error")
        assert pytest_wrapped_e.type == ValueError


def test_maximizer_objfun_failure_modes_unknown_metric():
    """Test for maximize-minimize failure mode:
    Use of unknown objfun
    """
    with pytest.raises(NotImplementedError) as pytest_wrapped_e:
        _ = _get_objfun_minimize_or_maximize(obj_func="unknown_of")
        assert pytest_wrapped_e.type == NotImplementedError



"""Module to compute climate change statistics using xscen functions."""

from typing import Optional, Union

import numpy as np
import xarray as xr

# Special imports from xscen
from xscen import climatological_op, compute_deltas, ensemble_stats, produce_horizon

__all__ = [
    "climatological_op",
    "compute_deltas",
    "ensemble_stats",
    "produce_horizon",
    "sampled_indicators",
]


def sampled_indicators(
    ds: xr.Dataset,
    deltas: xr.Dataset,
    delta_type: str,
    *,
    ds_weights: Optional[xr.DataArray] = None,
    delta_weights: Optional[xr.DataArray] = None,
    n: int = 50000,
    seed: int = None,
    return_dist: bool = False,
) -> Union[xr.Dataset, tuple[xr.Dataset, xr.Dataset, xr.Dataset, xr.Dataset]]:
    """Compute future indicators using a perturbation approach and random sampling.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset containing the historical indicators. The indicators are expected to be represented by a distribution of pre-computed percentiles.
        The percentiles should be stored in either a dimension called "percentile" [0, 100] or "quantile" [0, 1].
    deltas : xr.Dataset
        Dataset containing the future deltas to apply to the historical indicators.
    delta_type : str
        Type of delta provided. Must be one of ['absolute', 'percentage'].
    ds_weights : xr.DataArray, optional
        Weights to use when sampling the historical indicators, for dimensions other than 'percentile'/'quantile'.
        Dimensions not present in this Dataset, or if None, will be sampled uniformly unless they are shared with 'deltas'.
    delta_weights : xr.DataArray, optional
        Weights to use when sampling the deltas, such as along the 'realization' dimension.
        Dimensions not present in this Dataset, or if None, will be sampled uniformly unless they are shared with 'ds'.
    n : int
        Number of samples to generate.
    seed : int
        Seed to use for the random number generator.
    return_dist : bool
        Whether to return the full distributions (ds, deltas, fut) or only the percentiles.

    Returns
    -------
    fut_pct : xr.Dataset
        Dataset containing the future percentiles.
    ds_dist : xr.Dataset
        The historical distribution, stacked along the 'sample' dimension.
    deltas_dist : xr.Dataset
        The delta distribution, stacked along the 'sample' dimension.
    fut_dist : xr.Dataset
        The future distribution, stacked along the 'sample' dimension.

    Notes
    -----
    The future percentiles are computed as follows:
    1. Sample 'n' values from the historical distribution, weighting the percentiles by their associated coverage.
    2. Sample 'n' values from the delta distribution, using the provided weights.
    3. Create the future distribution by applying the sampled deltas to the sampled historical distribution, element-wise.
    4. Compute the percentiles of the future distribution.
    """
    # Prepare weights
    shared_dims = set(ds.dims).intersection(set(deltas.dims))
    exclude_dims = ["time", "horizon"]
    percentile_weights = _percentile_weights(ds)
    if ds_weights is not None:
        percentile_weights = (
            percentile_weights.expand_dims(
                {dim: ds_weights[dim] for dim in ds_weights.dims}
            )
            * ds_weights
        )
    percentile_weights = percentile_weights.expand_dims(
        {
            dim: ds[dim]
            for dim in set(ds.dims).difference(
                list(shared_dims) + list(percentile_weights.dims) + exclude_dims
            )
        }
    )
    if delta_weights is None:
        dims = set(deltas.dims).difference(list(shared_dims) + exclude_dims)
        delta_weights = xr.DataArray(
            np.ones([deltas.sizes[dim] for dim in dims]),
            coords={dim: deltas[dim] for dim in dims},
            dims=dims,
        )
    delta_weights = delta_weights.expand_dims(
        {
            dim: deltas[dim]
            for dim in set(deltas.dims).difference(
                list(shared_dims) + list(delta_weights.dims) + exclude_dims
            )
        }
    )

    unique_dims = set(percentile_weights.dims).symmetric_difference(
        set(delta_weights.dims)
    )
    if any([dim in shared_dims for dim in unique_dims]):
        problem_dims = [dim for dim in unique_dims if dim in shared_dims]
        raise ValueError(
            f"Dimension(s) {problem_dims} is shared between 'ds' and 'deltas', but not between 'ds_weights' and 'delta_weights'."
        )

    # Sample the distributions
    _, pdim, mult = _perc_or_quantile(ds)
    ds_dist = (
        _weighted_sampling(ds, percentile_weights, n=n, seed=seed)
        .drop_vars(["sample", *percentile_weights.dims])
        .assign_coords({"sample": np.arange(n)})
    )
    deltas_dist = (
        _weighted_sampling(deltas, delta_weights, n=n, seed=seed)
        .drop_vars(["sample", *delta_weights.dims])
        .assign_coords({"sample": np.arange(n)})
    )

    # Element-wise multiplication of the ref_dist and ens_dist
    if delta_type == "percentage":
        fut_dist = ds_dist * (1 + deltas_dist / 100)
    elif delta_type == "absolute":
        fut_dist = ds_dist + deltas_dist
    else:
        raise ValueError(
            f"Unknown operation '{delta_type}', expected one of ['absolute', 'percentage']."
        )

    # Compute future percentiles
    fut_pct = fut_dist.quantile(ds.percentile / mult, dim="sample")

    if pdim == "percentile":
        fut_pct = fut_pct.rename({"quantile": "percentile"})
        fut_pct["percentile"] = ds.percentile

    if return_dist:
        return fut_pct, ds_dist, deltas_dist, fut_dist
    else:
        return fut_pct


def _percentile_weights(da: Union[xr.DataArray, xr.Dataset]) -> xr.DataArray:
    """Compute the weights associated with percentiles, with support for unevenly spaced percentiles.

    Parameters
    ----------
    da : xr.DataArray or xr.Dataset
        DataArray or Dataset containing the percentiles to use when sampling.
        The percentiles are expected to be stored in either a dimension called "percentile" [0, 100] or "quantile" [0, 1].

    Returns
    -------
    p : xr.DataArray
        DataArray containing the weights associated with the percentiles.
    """
    pct, pdim, multiplier = _perc_or_quantile(da)

    # Temporarily add a 0 and 100th percentile
    p0 = xr.DataArray([0], coords={pdim: [0]}, dims=[pdim])
    p1 = xr.DataArray([multiplier], coords={pdim: [multiplier]}, dims=[pdim])
    p = xr.concat([p0, pct, p1], dim=pdim)
    p = p.diff(pdim) / 2
    p[0] = (
        p[0] * 2
    )  # The first and last weights need to be doubled to account for the fact that the first and last percentiles are not centered
    p[-1] = p[-1] * 2
    p = p.rolling({pdim: 2}, center=True).sum().shift({pdim: -1})[:-1]

    return p


def _weighted_sampling(
    ds: xr.Dataset, weights: xr.DataArray, n: int = 50000, seed: int = None
) -> xr.Dataset:
    """Sample from a distribution with weights.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset containing the distribution to sample from.
    weights : xr.DataArray
        Weights to use when sampling.
    n : int
        Number of samples to generate.
    seed : int
        Seed to use for the random number generator.

    Returns
    -------
    ds_dist : xr.Dataset
        Dataset containing the 'n' samples, stacked along the 'sample' dimension.
    """
    # Prepare the weights
    if weights.isnull().any():
        raise ValueError("The weights contain NaNs.")
    weights = weights / weights.sum()  # Must equal 1

    # For performance reasons, remove the chunking along impacted dimensions
    ds = ds.chunk({dim: -1 for dim in weights.dims})
    weights = weights.chunk({dim: -1 for dim in weights.dims})

    # Stack the dimensions containing weights
    ds = ds.stack({"sample": sorted(list(weights.dims))})
    weights = weights.stack({"sample": sorted(list(weights.dims))})
    weights = weights.reindex_like(ds)

    # Sample the dimensions with weights
    rng = np.random.default_rng(seed=seed)
    idx = rng.choice(weights.size, size=n, p=weights)

    # Create the distribution dataset
    ds_dist = ds.isel({"sample": idx})

    return ds_dist


def _perc_or_quantile(
    da: Union[xr.DataArray, xr.Dataset]
) -> tuple[xr.DataArray, str, int]:
    """Return 'percentile' or 'quantile' depending on the name of the percentile dimension."""
    if isinstance(da, xr.DataArray):
        if len(da.dims) != 1:
            raise ValueError(
                f"DataArray has more than one dimension: received {da.dims}."
            )
        pdim = str(da.dims[0])
        pct = da
        if pdim not in ["percentile", "quantile"]:
            raise ValueError(
                f"DataArray has no 'percentile' or 'quantile' dimension: received {pdim}."
            )
    else:
        pdim = [dim for dim in da.dims if dim in ["percentile", "quantile"]]
        if len(pdim) != 1:
            raise ValueError(
                "The Dataset should contain one of ['percentile', 'quantile']."
            )
        pdim = pdim[0]
        pct = da[pdim]

    multiplier = 100 if pdim == "percentile" else 1
    if (pct.min() < 0 or pct.max() > multiplier) or (
        pdim == "percentile" and pct.max() <= 1
    ):
        raise ValueError(
            f"The {pdim} values do not seem to be in the [0, {multiplier}] range."
        )

    return pct, pdim, multiplier



"""Hydrological analysis library built with xarray."""

from . import cc, indicators, utils

# Import top-level functions
# TODO: Decide which functions to import at the top level

__author__ = """Thomas-Charles Fortier Filion"""
__email__ = "tcff_hydro@outlook.com"
__version__ = "0.3.4"



import numpy as np
import pytest
import xarray as xr

import xhydro as xh


# Smoke test for xscen functions that are imported into xhydro
def test_xscen_imported():
    assert all(
        callable(getattr(xh.cc, f))
        for f in [
            "climatological_op",
            "compute_deltas",
            "ensemble_stats",
            "produce_horizon",
        ]
    )


class TestSampledIndicators:
    @pytest.mark.parametrize("delta_type", ["absolute", "percentage", "foo"])
    def test_sampled_indicators_type(self, delta_type):
        ds = xr.DataArray(
            np.arange(1, 8), coords={"percentile": [1, 10, 20, 50, 80, 90, 99]}
        ).to_dataset(name="QMOYAN")
        deltas = xr.DataArray(
            np.arange(-10, 55, 5),
            coords={"realization": np.arange(13)},
        ).to_dataset(name="QMOYAN")
        if delta_type in ["absolute", "percentage"]:
            out = xh.cc.sampled_indicators(
                ds, deltas, delta_type, n=10, seed=42, return_dist=True
            )

            np.testing.assert_array_equal(
                out[0]["percentile"], [1, 10, 20, 50, 80, 90, 99]
            )
            assert all(
                chosen in np.arange(1, 8) for chosen in np.unique(out[1].QMOYAN.values)
            )
            assert all(
                chosen in np.arange(-10, 55, 5)
                for chosen in np.unique(out[2].QMOYAN.values)
            )

            if delta_type == "absolute":
                assert (
                    np.min(out[3].QMOYAN) >= 1 - 10
                )  # Min of historical minus min of deltas
                assert (
                    np.max(out[3].QMOYAN) <= 7 + 50
                )  # Max of historical plus max of deltas
                np.testing.assert_array_almost_equal(
                    out[0]["QMOYAN"].values, [-3.0, -3.0, 14.6, 40.0, 46.2, 51.6, 56.46]
                )
            else:
                assert np.min(out[3].QMOYAN) >= 1 * (
                    1 - 10 / 100
                )  # Min of historical minus min of deltas
                assert np.max(out[3].QMOYAN) <= 7 * (
                    1 + 50 / 100
                )  # Max of historical plus max of deltas
                np.testing.assert_array_almost_equal(
                    out[0]["QMOYAN"].values, [1.9, 1.9, 4.06, 6.75, 7.34, 8.88, 10.338]
                )

        else:
            with pytest.raises(
                ValueError, match=f"Unknown operation '{delta_type}', expected one"
            ):
                xh.cc.sampled_indicators(ds, deltas, delta_type, n=10, seed=42)

    def test_sampled_indicators_return(self):
        ds = xr.DataArray(
            np.arange(1, 8), coords={"percentile": [1, 10, 20, 50, 80, 90, 99]}
        ).to_dataset(name="QMOYAN")
        deltas = xr.DataArray(
            np.arange(-10, 55, 5),
            coords={"realization": np.arange(13)},
        ).to_dataset(name="QMOYAN")

        out1 = xh.cc.sampled_indicators(
            ds, deltas, "absolute", n=10, seed=42, return_dist=False
        )
        assert isinstance(out1, xr.Dataset)

        out2 = xh.cc.sampled_indicators(
            ds, deltas, "absolute", n=10, seed=42, return_dist=True
        )
        assert isinstance(out2, tuple)
        assert len(out2) == 4
        assert all(isinstance(o, xr.Dataset) for o in out2)
        assert out2[0].equals(out1)
        for o in out2[1:]:
            assert list(o.dims) == ["sample"]
            assert len(o["sample"]) == 10

    @pytest.mark.parametrize("weights", [None, "ds", "deltas", "both"])
    def test_sampled_indicators_weights(self, weights):
        ds = xr.DataArray(
            np.array(
                [
                    [[1, 2, 3, 4, 5], [101, 102, 103, 104, 105]],
                    [[6, 7, 8, 9, 10], [106, 107, 108, 109, 110]],
                ]
            ),
            dims=("foo", "station", "percentile"),
            coords={
                "percentile": [1, 25, 50, 75, 99],
                "station": ["a", "b"],
                "foo": ["bar", "baz"],
            },
        ).to_dataset(name="QMOYAN")
        deltas = xr.DataArray(
            np.array(
                [
                    [[-10, -5, 0, 5, 10], [-25, -20, -15, -10, -5]],
                    [[40, 45, 50, 55, 60], [115, 120, 125, 130, 135]],
                ]
            ),
            dims=("platform", "station", "realization"),
            coords={
                "realization": [1, 25, 50, 75, 99],
                "station": ["a", "b"],
                "platform": ["x", "y"],
            },
        ).to_dataset(name="QMOYAN")
        ds_weights = xr.DataArray(
            np.array([0, 1]),
            dims="foo",
            coords={
                "foo": ["bar", "baz"],
            },
        )
        delta_weights = xr.DataArray(
            np.array([[0, 0, 0, 1, 0], [0, 0, 5, 0, 0]]),
            dims=("platform", "realization"),
            coords={
                "realization": [1, 25, 50, 75, 99],
                "platform": ["x", "y"],
            },
        )

        expected_out1 = (
            np.arange(1, 11)
            if (weights is None or weights == "deltas")
            else np.arange(6, 11)
        )
        expected_out2 = (
            [-10, -5, 0, 5, 10, 40, 45, 50, 55, 60]
            if (weights is None or weights == "ds")
            else [5, 50]
        )

        if weights is None:
            out = xh.cc.sampled_indicators(
                ds, deltas, "percentage", n=10, seed=42, return_dist=True
            )
            np.testing.assert_array_almost_equal(
                out[0].QMOYAN.isel(station=0).values, [1.809, 5.5, 11.8, 12.0, 15.8155]
            )
        elif weights == "ds":
            out = xh.cc.sampled_indicators(
                ds,
                deltas,
                "percentage",
                n=10,
                seed=42,
                return_dist=True,
                ds_weights=ds_weights,
            )
            np.testing.assert_array_almost_equal(
                out[0].QMOYAN.isel(station=0).values,
                [5.427, 8.8, 13.275, 13.5, 15.8155],
            )
        elif weights == "deltas":
            out = xh.cc.sampled_indicators(
                ds,
                deltas,
                "percentage",
                n=10,
                seed=42,
                return_dist=True,
                delta_weights=delta_weights,
            )
            np.testing.assert_array_almost_equal(
                out[0].QMOYAN.isel(station=0).values, [2.1, 7.5, 12.0, 12.0, 14.865]
            )
            assert sum(out[2].QMOYAN.isel(station=0).values == 5) < sum(
                out[2].QMOYAN.isel(station=0).values == 50
            )  # 50 should be sampled more often
        elif weights == "both":
            out = xh.cc.sampled_indicators(
                ds,
                deltas,
                "percentage",
                n=10,
                seed=42,
                return_dist=True,
                ds_weights=ds_weights,
                delta_weights=delta_weights,
            )
            np.testing.assert_array_almost_equal(
                out[0].QMOYAN.isel(station=0).values, [6.3, 12.0, 13.5, 13.5, 14.865]
            )

        assert all(
            chosen in expected_out1
            for chosen in np.unique(out[1].QMOYAN.isel(station=0).values)
        )
        assert all(
            chosen in expected_out2
            for chosen in np.unique(out[2].QMOYAN.isel(station=0).values)
        )
        assert all(
            "station" in o.dims for o in out
        )  # "station" is a shared dimension, so it should be in all outputs

    def test_sampled_indicators_weight_err(self):
        ds = xr.DataArray(
            np.array(
                [
                    [[1, 2, 3, 4, 5], [101, 102, 103, 104, 105]],
                    [[6, 7, 8, 9, 10], [106, 107, 108, 109, 110]],
                ]
            ),
            dims=("platform", "station", "percentile"),
            coords={
                "percentile": [1, 25, 50, 75, 99],
                "station": ["a", "b"],
                "platform": ["x", "y"],
            },
        ).to_dataset(name="QMOYAN")
        deltas = xr.DataArray(
            np.array(
                [
                    [[-10, -5, 0, 5, 10], [-25, -20, -15, -10, -5]],
                    [[40, 45, 50, 55, 60], [115, 120, 125, 130, 135]],
                ]
            ),
            dims=("platform", "station", "realization"),
            coords={
                "realization": [1, 25, 50, 75, 99],
                "station": ["a", "b"],
                "platform": ["x", "y"],
            },
        ).to_dataset(name="QMOYAN")
        delta_weights = xr.DataArray(
            np.array([[0, 0, 0, 1, 0], [0, 0, 5, 0, 0]]),
            dims=("platform", "realization"),
            coords={
                "realization": [1, 25, 50, 75, 99],
                "platform": ["x", "y"],
            },
        )

        with pytest.raises(
            ValueError,
            match="is shared between 'ds' and 'deltas', but not between 'ds_weights' and 'delta_weights'.",
        ):
            xh.cc.sampled_indicators(
                ds, deltas, "percentage", n=10, seed=42, delta_weights=delta_weights
            )

    def test_p_weights(self):
        def _make_da(arr):
            return xr.DataArray(arr, dims="percentile", coords={"percentile": arr})

        out = xh.cc._percentile_weights(_make_da(np.linspace(0, 100, 101)))
        ans = np.ones(101)
        ans[0] = 0.5
        ans[-1] = 0.5
        np.testing.assert_array_equal(out, ans)

        out = xh.cc._percentile_weights(
            _make_da(np.array([1, 10, 20, 30, 40, 50, 60, 70, 80, 90, 99]))
        )
        ans = np.array([5.5, 9.5, 10, 10, 10, 10, 10, 10, 10, 9.5, 5.5])
        np.testing.assert_array_equal(out, ans)

    @pytest.mark.parametrize("weights_ndim", [1, 2])
    def test_weighted_sampling(self, weights_ndim):
        data = np.array([[[1, 2, 3], [101, 102, 103]], [[2, 3, 4], [102, 103, 104]]])
        ds = xr.DataArray(
            data,
            dims=("platform", "station", "percentile"),
            coords={
                "percentile": [25, 50, 75],
                "station": ["a", "b"],
                "platform": ["x", "y"],
            },
        ).to_dataset(name="QMOYAN")
        weights = xr.DataArray(
            np.array([[15, 100, 33], [500, 0, 0]]).T,
            dims=("percentile", "platform"),
            coords={
                "platform": ["y", "x"],  # Invert order to test reordering
                "percentile": [25, 50, 75],
            },
        )
        if weights_ndim == 1:
            weights = weights.isel(platform=0)

        out = xh.cc._weighted_sampling(ds, weights, n=10, seed=42)
        assert set(out.dims) == set(
            ["station", "sample"] + (["platform"] if weights_ndim == 1 else [])
        )
        assert list(out["sample"].coords) == ["sample", "percentile"] + (
            ["platform"] if weights_ndim == 2 else []
        )
        assert len(out["sample"]) == 10

        np.testing.assert_array_equal(
            out["percentile"],
            {
                1: [50, 50, 75, 50, 25, 75, 50, 75, 50, 50],
                2: [25, 25, 50, 25, 25, 75, 25, 25, 25, 25],
            }[weights_ndim],
        )  # With seed=42
        if weights_ndim == 2:
            np.testing.assert_array_equal(
                out["platform"], ["y", "x", "y", "x", "x", "y", "x", "y", "x", "x"]
            )

    def test_weighted_sampling_errors(self):
        ds = xr.DataArray(
            np.arange(10),
        ).to_dataset(
            name="foo"
        )  # Doesn't matter what the data is here
        weights = xr.DataArray(
            [[1, 2, 3], [4, np.nan, 6]],
        )
        with pytest.raises(ValueError, match="The weights contain NaNs."):
            xh.cc._weighted_sampling(ds, weights, n=10, seed=42)

    @pytest.mark.parametrize("dim", ["percentile", "quantile"])
    def test_p_or_q(self, dim):
        if dim == "percentile":
            da = xr.DataArray(
                np.linspace(0, 100, 101),
                dims="percentile",
                attrs={"units": "%"},
            )
            da = da.assign_coords(percentile=da)
        else:
            da = xr.DataArray(np.linspace(0, 1, 101), dims="quantile")
            da = da.assign_coords(quantile=da)
        ds = da.to_dataset(name="foo").expand_dims({"time": np.arange(10)})
        out_da, pdim_da, mult_da = xh.cc._perc_or_quantile(da)
        out_ds, pdim_ds, mult_ds = xh.cc._perc_or_quantile(ds)

        assert out_da.equals(da)
        assert out_ds.equals(ds[dim])
        assert pdim_da == pdim_ds == dim
        assert mult_da == mult_ds == (100 if dim == "percentile" else 1)

    def test_p_or_q_errors(self):
        # Test 1: DataArray with >1 dimension
        da = xr.DataArray(
            np.linspace(0, 1, 101),
            dims="quantile",
        ).expand_dims({"time": np.arange(10)})
        with pytest.raises(ValueError, match="DataArray has more than one dimension"):
            xh.cc._perc_or_quantile(da)

        # Test 2: DataArray/Datset with no percentile or quantile dimension
        da = xr.DataArray(
            np.linspace(0, 1, 101),
            dims="foo",
        )
        with pytest.raises(
            ValueError, match="DataArray has no 'percentile' or 'quantile' dimension"
        ):
            xh.cc._perc_or_quantile(da)
        ds = da.to_dataset(name="foo")
        with pytest.raises(ValueError, match="The Dataset should contain one of "):
            xh.cc._perc_or_quantile(ds)

        # Wrong range
        da = xr.DataArray(
            np.linspace(0, 2, 101),
            dims="quantile",
        )
        with pytest.raises(ValueError, match="values do not seem to be in the "):
            xh.cc._perc_or_quantile(da)
        da = xr.DataArray(
            np.linspace(-1, 1, 101),
            dims="quantile",
        )
        with pytest.raises(ValueError, match="values do not seem to be in the "):
            xh.cc._perc_or_quantile(da)
        da = xr.DataArray(
            np.linspace(0, 1, 101),
            dims="percentile",
        )
        with pytest.raises(ValueError, match="values do not seem to be in the "):
            xh.cc._perc_or_quantile(da)



"""Helper functions for testing data management."""

import importlib.resources as ilr
import logging
import os
from pathlib import Path
from typing import Optional, Union
from urllib.parse import urljoin

import pooch

from xhydro import __version__ as __xhydro_version__

__all__ = [
    "DATA_DIR",
    "DATA_URL",
    "DEVEREAUX",
    "generate_registry",
    "load_registry",
    "populate_testing_data",
]

_default_cache_dir = pooch.os_cache("xhydro-testdata")

DATA_DIR = os.getenv("XHYDRO_DATA_DIR", _default_cache_dir)
"""Sets the directory to store the testing datasets.

If not set, the default location will be used (based on ``platformdirs``, see :func:`pooch.os_cache`).

Notes
-----
When running tests locally, this can be set for both `pytest` and `tox` by exporting the variable:

.. code-block:: console

    $ export XHYDRO_DATA_DIR="/path/to/my/data"

or setting the variable at runtime:

.. code-block:: console

    $ env XHYDRO_DATA_DIR="/path/to/my/data" pytest
"""

TESTDATA_BRANCH = os.getenv("XHYDRO_TESTDATA_BRANCH", "main")
"""Sets the branch of hydrologie/xhydro-testdata to use when fetching testing datasets.

Notes
-----
When running tests locally, this can be set for both `pytest` and `tox` by exporting the variable:

.. code-block:: console

    $ export XHYDRO_TESTDATA_BRANCH="my_testing_branch"

or setting the variable at runtime:

.. code-block:: console

    $ env XHYDRO_TESTDATA_BRANCH="my_testing_branch" pytest
"""

DATA_URL = f"https://github.com/hydrologie/xhydro-testdata/raw/{TESTDATA_BRANCH}"


def generate_registry(
    filenames: Optional[list[str]] = None, base_url: str = DATA_URL
) -> None:
    """Generate a registry file for the test data.

    Parameters
    ----------
    filenames : list of str, optional
        List of filenames to generate the registry file for.
        If not provided, all files under xhydro/testing/data will be used.
    base_url : str, optional
        Base URL to the test data repository.
    """
    # Gather the data folder and registry file locations from installed package_data
    data_folder = ilr.files("xhydro").joinpath("testing/data")
    registry_file = ilr.files("xhydro").joinpath("testing/registry.txt")

    # Download the files to the installed xhydro/testing/data folder
    if filenames is None:
        with ilr.as_file(data_folder) as data:
            for file in data.rglob("*"):
                filename = file.relative_to(data).as_posix()
                pooch.retrieve(
                    url=urljoin(base_url, filename),
                    known_hash=None,
                    fname=filename,
                    path=data_folder,
                )

    # Generate the registry file
    with ilr.as_file(data_folder) as data, ilr.as_file(registry_file) as registry:
        pooch.make_registry(data.as_posix(), registry.as_posix())


def load_registry(file: Optional[Union[str, Path]] = None) -> dict[str, str]:
    """Load the registry file for the test data.

    Parameters
    ----------
    file : str or Path, optional
        Path to the registry file. If not provided, the registry file found within the package data will be used.

    Returns
    -------
    dict
        Dictionary of filenames and hashes.
    """
    # Get registry file from package_data
    if file is None:
        registry_file = ilr.files("xhydro").joinpath("testing/registry.txt")
        if registry_file.is_file():
            logging.info("Registry file found in package_data: %s", registry_file)
    else:
        registry_file = Path(file)
        if not registry_file.is_file():
            raise FileNotFoundError(f"Registry file not found: {registry_file}")

    # Load the registry file
    registry = dict()
    with registry_file.open() as buffer:
        for entry in buffer.readlines():
            registry[entry.split()[0]] = entry.split()[1]

    return registry


DEVEREAUX = pooch.create(
    path=pooch.os_cache("xhydro-testdata"),
    base_url=DATA_URL,
    version=__xhydro_version__,
    version_dev="main",
    env="XHYDRO_DATA_DIR",
    allow_updates="XHYDRO_DATA_UPDATES",
    registry=load_registry(),
)
"""Pooch registry instance for xhydro test data.

Notes
-----
There are two environment variables that can be used to control the behaviour of this registry:

  - ``XHYDRO_DATA_DIR``: If this environment variable is set, it will be used as the base directory to store the data
    files. The directory should be an absolute path (i.e., it should start with ``/``). Otherwise,
    the default location will be used (based on ``platformdirs``, see :func:`pooch.os_cache`).

  - ``XHYDRO_DATA_UPDATES``: If this environment variable is set, then the data files will be downloaded even if the
    upstream hashes do not match. This is useful if you want to always use the latest version of the data files.

Examples
--------
Using the registry to download a file:

.. code-block:: python

    from xhydro.testing.utils import DEVEREAUX
    import xarray as xr

    example_file = DEVEREAUX.fetch("example.nc")
    data = xr.open_dataset(example_file)
"""


def populate_testing_data(
    registry: Optional[Union[str, Path]] = None,
    temp_folder: Optional[Path] = None,
    branch: str = TESTDATA_BRANCH,
    _local_cache: Path = _default_cache_dir,
) -> None:
    """Populate the local cache with the testing data.

    Parameters
    ----------
    registry : str or Path, optional
        Path to the registry file. If not provided, the registry file from package_data will be used.
    temp_folder : Path, optional
        Path to a temporary folder to use as the local cache. If not provided, the default location will be used.
    branch : str, optional
        Branch of hydrologie/xhydro-testdata to use when fetching testing datasets.
    _local_cache : Path, optional
        Path to the local cache. Defaults to the default location.

    Returns
    -------
    None
        The testing data will be downloaded to the local cache.
    """
    # Get registry file from package_data or provided path
    registry = load_registry(registry)

    # Set the local cache to the temp folder
    if temp_folder is not None:
        _local_cache = temp_folder
    # Set the branch
    DEVEREAUX.version_dev = branch
    # Set the local cache
    DEVEREAUX.path = _local_cache

    # Download the files
    for filename in registry.keys():
        DEVEREAUX.fetch(filename)



"""Module to compute indicators using xclim's build_indicator_module_from_yaml."""

import warnings
from typing import Optional

import xarray as xr
import xclim as xc
import xscen as xs
from xclim.core.units import rate2amount

# Special imports from xscen
from xscen import compute_indicators

__all__ = [
    "compute_indicators",
    "compute_volume",
    "get_yearly_op",
]


def compute_volume(
    da: xr.DataArray, *, out_units: str = "m3", attrs: Optional[dict] = None
) -> xr.DataArray:
    """Compute the volume of water from a streamflow variable, keeping the same frequency.

    Parameters
    ----------
    da : xr.DataArray
        Streamflow variable.
    out_units : str
        Output units. Defaults to "m3".
    attrs : dict, optional
        Attributes to add to the output variable.
        Default attributes for "long_name", "units", "cell_methods" and "description" will be added if not provided.

    Returns
    -------
    xr.DataArray
        Volume of water.
    """
    default_attrs = {
        "long_name": "Volume of water",
        "cell_methods": "time: sum",
        "description": "Volume of water",
    }
    attrs = attrs or {}
    # Add default attributes
    for k, v in default_attrs.items():
        attrs.setdefault(k, v)

    out = rate2amount(da, out_units=out_units)
    out.attrs.update(attrs)

    return out


def get_yearly_op(
    ds,
    op,
    *,
    input_var: str = "streamflow",
    window: int = 1,
    timeargs: Optional[dict] = None,
    missing: str = "skip",
    missing_options: Optional[dict] = None,
    interpolate_na: bool = False,
) -> xr.Dataset:
    """Compute yearly operations on a variable.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset containing the variable to compute the operation on.
    op : str
        Operation to compute. One of ["max", "min", "mean", "sum"].
    input_var : str
        Name of the input variable. Defaults to "streamflow".
    window : int
        Size of the rolling window. A "mean" operation is performed on the rolling window before the call to xclim.
        This parameter cannot be used with the "sum" operation.
    timeargs : dict, optional
        Dictionary of time arguments for the operation.
        Keys are the name of the period that will be added to the results (e.g. "winter", "summer", "annual").
        Values are up to two dictionaries, with both being optional.
        The first is {'freq': str}, where str is a frequency supported by xarray (e.g. "YS", "YS-JAN", "YS-DEC").
        It needs to be a yearly frequency. Defaults to "YS-JAN".
        The second is an indexer as supported by :py:func:`xclim.core.calendar.select_time`.
        Defaults to {}, which means the whole year.
        See :py:func:`xclim.core.calendar.select_time` for more information.
        Examples: {"winter": {"freq": "YS-DEC", "date_bounds": ["12-01", "02-28"]}}, {"jan": {"freq": "YS", "month": 1}}, {"annual": {}}.
    missing : str
        How to handle missing values. One of "skip", "any", "at_least_n", "pct", "wmo".
        See :py:func:`xclim.core.missing` for more information.
    missing_options : dict, optional
        Dictionary of options for the missing values' method. See :py:func:`xclim.core.missing` for more information.
    interpolate_na : bool
        Whether to interpolate missing values before computing the operation. Only used with the "sum" operation.
        Defaults to False.

    Returns
    -------
    xr.Dataset
        Dataset containing the computed operations, with one variable per indexer.
        The name of the variable follows the pattern `{input_var}{window}_{op}_{indexer}`.

    Notes
    -----
    If you want to perform a frequency analysis on a frequency that is finer than annual, simply use multiple timeargs
    (e.g. 1 per month) to create multiple distinct variables.
    """
    missing_options = missing_options or {}
    timeargs = timeargs or {"annual": {}}

    if op not in ["max", "min", "mean", "sum"]:
        raise ValueError(
            f"Operation {op} is not supported. Please use one of ['max', 'min', 'mean', 'sum']."
        )
    if op == "sum":
        if window > 1:
            raise ValueError("Cannot use a rolling window with a sum operation.")
        if interpolate_na:
            ds[input_var] = ds[input_var].interpolate_na(dim="time", method="linear")

    # Add the variable to xclim to avoid raising an error
    if input_var not in xc.core.utils.VARIABLES:
        attrs = {
            "long_name": None,
            "units": None,
            "cell_methods": None,
            "description": None,
        }
        attrs.update(ds[input_var].attrs)
        attrs["canonical_units"] = attrs["units"]
        attrs.pop("units")
        xc.core.utils.VARIABLES[input_var] = attrs

    # FIXME: This should be handled by xclim once it supports rolling stats (Issue #1480)
    # rolling window
    if window > 1:
        ds[input_var] = (
            ds[input_var]
            .rolling(dim={"time": window}, min_periods=window, center=False)
            .mean()
        )

    indicators = []
    month_labels = [
        "JAN",
        "FEB",
        "MAR",
        "APR",
        "MAY",
        "JUN",
        "JUL",
        "AUG",
        "SEP",
        "OCT",
        "NOV",
        "DEC",
    ]
    for i in timeargs:
        freq = timeargs[i].get("freq", "YS-JAN")
        if not xc.core.calendar.compare_offsets(freq, "==", "YS"):
            raise ValueError(
                f"Frequency {freq} is not supported. Please use a yearly frequency."
            )
        indexer = {k: v for k, v in timeargs[i].items() if k != "freq"}
        if len(indexer) > 1:
            raise ValueError("Only one indexer is supported per operation.")

        # Manage the frequency
        if (
            "season" in indexer.keys()
            and "DJF" in indexer["season"]
            and freq != "YS-DEC"
        ):
            warnings.warn(
                "The frequency is not YS-DEC, but the season indexer includes DJF. "
                "This will lead to misleading results."
            )
        elif (
            "doy_bounds" in indexer.keys()
            and indexer["doy_bounds"][0] >= indexer["doy_bounds"][1]
        ) or (
            "date_bounds" in indexer.keys()
            and int(indexer["date_bounds"][0].split("-")[0])
            >= int(indexer["date_bounds"][1].split("-")[0])
        ):
            if "doy_bounds" in indexer.keys():
                # transform doy to a date to find the month
                ts = xr.cftime_range(
                    start="2000-01-01",
                    periods=366,
                    freq="D",
                    calendar=ds.time.dt.calendar,
                )
                month_start = ts[indexer["doy_bounds"][0] - 1].month
                month_end = ts[indexer["doy_bounds"][1] - 1].month
            else:
                month_start = int(indexer["date_bounds"][0].split("-")[0])
                month_end = int(indexer["date_bounds"][1].split("-")[0])
            if month_end == month_start:
                warnings.warn(
                    "The bounds wrap around the year, but the month is the same between the both of them. "
                    "This is not supported and will lead to wrong results."
                )
            if freq == "YS" or (month_start != month_labels.index(freq.split("-")[1])):
                warnings.warn(
                    f"The frequency is {freq}, but the bounds are between months {month_start} and {month_end}. "
                    f"You should use 'YS-{month_labels[month_start - 1]}' as the frequency."
                )

        identifier = f"{input_var}{window if window > 1 else ''}_{op}_{i.lower()}"
        ind = xc.core.indicator.Indicator.from_dict(
            data={
                "base": "stats",
                "input": {"da": input_var},
                "parameters": {
                    "op": op if op != "sum" else "integral",
                    "indexer": indexer,
                    "freq": freq,
                },
                "missing": missing,
                "missing_options": missing_options,
            },
            identifier=identifier,
            module="fa",
        )
        indicators.append((identifier, ind))

    # Compute the indicators
    ind_dict = compute_indicators(ds, indicators=indicators)

    # Combine all the indicators into one dataset
    out = xr.merge(
        [
            da.assign_coords(
                time=xr.date_range(
                    da.time[0].dt.strftime("%Y-01-01").item(),
                    periods=da.time.size,
                    calendar=da.time.dt.calendar,
                    freq="YS",
                )
            )
            for da in ind_dict.values()
        ]
    )
    out = xs.clean_up(out, common_attrs_only=ind_dict)

    return out



import numpy as np
import pytest
import xarray as xr
from xclim.testing.helpers import test_timeseries as timeseries

import xhydro.frequency_analysis as xhfa
from xhydro.frequency_analysis.local import _get_plotting_positions, _prepare_plots


class TestFit:
    def test_fit(self):
        ds = timeseries(
            np.array([50, 65, 80, 95, 110, 125, 140, 155, 170, 185, 200]),
            variable="streamflow",
            start="2001-01-01",
            freq="YS",
            as_dataset=True,
        )
        params = xhfa.local.fit(ds, distributions=["gamma", "pearson3"])

        np.testing.assert_array_equal(params.dparams, ["a", "skew", "loc", "scale"])
        np.testing.assert_array_equal(params.scipy_dist, ["gamma", "pearson3"])
        assert params.streamflow.attrs["scipy_dist"] == ["gamma", "pearson3"]
        assert params.streamflow.attrs["estimator"] == "Maximum likelihood"
        assert params.streamflow.attrs["long_name"] == "Distribution parameters"
        assert (
            params.streamflow.attrs["description"] == "Parameters of the distributions"
        )

        np.testing.assert_array_almost_equal(
            params.streamflow,
            np.array(
                [
                    [9.95357815e00, np.nan, -3.07846650e01, 1.56498193e01],
                    [np.nan, -2.25674044e-05, 1.25012261e02, 4.74238877e01],
                ]
            ),
        )

    def test_default(self):
        ds = timeseries(
            np.array([50, 65, 80, 95, 110, 125, 140, 155, 170, 185, 200]),
            variable="streamflow",
            start="2001-01-01",
            freq="YS",
            as_dataset=True,
        )
        params = xhfa.local.fit(ds)
        np.testing.assert_array_equal(
            params.dparams, ["a", "c", "skew", "loc", "scale"]
        )
        np.testing.assert_array_equal(
            params.scipy_dist,
            [
                "expon",
                "gamma",
                "genextreme",
                "genpareto",
                "gumbel_r",
                "pearson3",
                "weibull_min",
            ],
        )

    @pytest.mark.parametrize("miny", [10, 15])
    def test_min_years(self, miny):
        ds = timeseries(
            np.array([50, 65, 80, 95, 110, 125, 140, 155, 170, 185, 200]),
            variable="streamflow",
            start="2001-01-01",
            freq="YS",
            as_dataset=True,
        )
        params = xhfa.local.fit(ds, distributions=["gamma"], min_years=miny)
        np.testing.assert_array_almost_equal(
            params.streamflow,
            (
                np.array(
                    [[9.95357815e00, -3.07846650e01, 1.56498193e01]]
                    if miny == 10
                    else [[np.nan, np.nan, np.nan]]
                )
            ),
        )


@pytest.mark.parametrize("mode", ["max", "min", "foo"])
def test_quantiles(mode):
    ds = timeseries(
        np.array([50, 65, 80, 95, 110, 125, 140, 155, 170, 185, 200]),
        variable="streamflow",
        start="2001-01-01",
        freq="YS",
        as_dataset=True,
    )
    params = xhfa.local.fit(ds, distributions=["gamma", "pearson3"])

    if mode == "foo":
        with pytest.raises(ValueError):
            xhfa.local.parametric_quantiles(params, [10, 20], mode=mode)
    else:
        rp = xhfa.local.parametric_quantiles(params, [10, 20], mode=mode)

        np.testing.assert_array_equal(rp.return_period, [10, 20])
        np.testing.assert_array_equal(
            rp.p_quantile, [0.1, 0.05] if mode == "min" else [0.9, 0.95]
        )
        np.testing.assert_array_equal(rp.scipy_dist, ["gamma", "pearson3"])
        assert rp.streamflow.attrs["long_name"] == "Return period"
        assert (
            rp.streamflow.attrs["description"]
            == f"Return period ({mode}) estimated with statistic distributions"
        )
        assert rp.streamflow.attrs["cell_methods"] == "dparams: ppf"
        assert rp.streamflow.attrs["mode"] == mode

        ans = np.array(
            [[190.66041057, 214.08102761], [185.78830382, 203.01731036]]
            if mode == "max"
            else [[66.00067153, 53.58658639], [64.23598869, 47.00660287]]
        )
        np.testing.assert_array_almost_equal(rp.streamflow, ans)


def test_criteria():
    ds = timeseries(
        np.array([50, 65, 80, 95, 110, 125, 140, 155, 170, 185, 200]),
        variable="streamflow",
        start="2001-01-01",
        freq="YS",
        as_dataset=True,
    )
    params = xhfa.local.fit(ds, distributions=["gamma", "pearson3"])
    crit = xhfa.local.criteria(ds, params)
    crit_with_otherdim = xhfa.local.criteria(
        ds.expand_dims("otherdim"), params.expand_dims("otherdim")
    )
    np.testing.assert_array_almost_equal(
        crit.streamflow, crit_with_otherdim.streamflow.squeeze("otherdim")
    )

    np.testing.assert_array_equal(crit.scipy_dist, ["gamma", "pearson3"])
    np.testing.assert_array_equal(crit.criterion, ["aic", "bic", "aicc"])
    assert crit.streamflow.attrs["long_name"] == "Information criteria"
    assert (
        crit.streamflow.attrs["description"]
        == "Information criteria for the distribution parameters."
    )
    assert crit.streamflow.attrs["scipy_dist"] == ["gamma", "pearson3"]
    assert all(
        attr not in crit.streamflow.attrs
        for attr in ["estimator", "method", "min_years"]
    )

    np.testing.assert_array_almost_equal(
        crit.streamflow,
        np.array(
            [
                [118.19066549, 118.58856076, 118.63510993],
                [118.12140939, 118.51930466, 118.56585383],
            ]
        ),
    )


class TestGetPlottingPositions:
    def test_default(self):
        data = timeseries(
            np.array([50, 65, 80, 95]),
            variable="streamflow",
            start="2001-01-01",
            freq="YS",
            as_dataset=True,
        )
        expected = np.array([1.16666667, 1.61538462, 2.625, 7.0])
        result = _get_plotting_positions(data)
        np.testing.assert_array_almost_equal(result.streamflow_pp, expected)
        np.testing.assert_array_almost_equal(result.streamflow, data.streamflow)

        data_2d = xr.concat([data, data], dim="id")
        result = _get_plotting_positions(data_2d)
        np.testing.assert_array_almost_equal(result.streamflow_pp, [expected, expected])
        np.testing.assert_array_equal(result.streamflow, data_2d.streamflow)

    def test_nan(self):
        data = timeseries(
            np.array([50, np.nan, 80, 95]),
            variable="streamflow",
            start="2001-01-01",
            freq="YS",
            as_dataset=True,
        )
        expected = np.array([1.23076923, np.nan, 2.0, 5.33333333])
        result = _get_plotting_positions(data)
        np.testing.assert_array_almost_equal(result.streamflow_pp, expected)
        np.testing.assert_array_almost_equal(result.streamflow, data.streamflow)

        data_2d = xr.concat([data, data], dim="id")
        result = _get_plotting_positions(data_2d)
        np.testing.assert_array_almost_equal(result.streamflow_pp, [expected, expected])
        np.testing.assert_array_equal(result.streamflow, data_2d.streamflow)

    def test_return_period(self):
        data = timeseries(
            np.array([50, 65, 80, 95]),
            variable="streamflow",
            start="2001-01-01",
            freq="YS",
            as_dataset=True,
        )
        expected = np.array([0.14285714, 0.38095238, 0.61904762, 0.85714286])
        result = _get_plotting_positions(data, return_period=False)
        np.testing.assert_array_almost_equal(result.streamflow_pp, expected)
        np.testing.assert_array_almost_equal(result.streamflow, data.streamflow)

        data_2d = xr.concat([data, data], dim="id")
        result = _get_plotting_positions(data_2d, return_period=False)
        np.testing.assert_array_almost_equal(result.streamflow_pp, [expected, expected])
        np.testing.assert_array_equal(result.streamflow, data_2d.streamflow)

    def test_alpha_beta(self):
        data = timeseries(
            np.array([50, 65, 80, 95]),
            variable="streamflow",
            start="2001-01-01",
            freq="YS",
            as_dataset=True,
        )
        expected = np.array([1.25, 1.66666667, 2.5, 5.0])
        result = _get_plotting_positions(data, alpha=0, beta=0)
        np.testing.assert_array_almost_equal(result.streamflow_pp, expected)
        np.testing.assert_array_almost_equal(result.streamflow, data.streamflow)

        data_2d = xr.concat([data, data], dim="id")
        result = _get_plotting_positions(data_2d, alpha=0, beta=0)
        np.testing.assert_array_almost_equal(result.streamflow_pp, [expected, expected])
        np.testing.assert_array_equal(result.streamflow, data_2d.streamflow)


class Testprepare_plots:
    ds = timeseries(
        np.array([50, 65, 80, 95, 110, 125, 140, 155, 170, 185, 200]),
        variable="streamflow",
        start="2001-01-01",
        freq="YS",
        as_dataset=True,
    )
    params = xhfa.local.fit(ds, distributions=["gamma", "pearson3"])

    def test_prepare_plots_default(self):
        result = _prepare_plots(self.params)
        assert result.streamflow.shape == (2, 100)
        assert result.return_period.min() == 1
        assert result.return_period.max() == 10000
        expected = np.array(
            [
                [-30.78466504, 63.64266447, 78.19229358, 88.62699148, 97.15369501],
                [-np.inf, 61.08708903, 79.72126025, 92.05411647, 101.59650405],
            ]
        )
        np.testing.assert_array_almost_equal(result.streamflow.head(), expected)

    def test_prepare_plots_linear(self):
        result = _prepare_plots(self.params, log=False)

        expected = np.array(
            [
                [-30.78466504, 262.72577254, 281.56238078, 292.27851004, 299.75980757],
                [-np.inf, 235.6878466, 247.41104463, 253.8825982, 258.32114457],
            ]
        )
        np.testing.assert_array_almost_equal(result.streamflow.head(), expected)

    def test_prepare_plots_range(self):
        result = _prepare_plots(self.params, xmin=5, xmax=500)
        assert result.return_period.min() == pytest.approx(5)
        assert result.return_period.max() == pytest.approx(500)

    def test_prepare_plots_npoints(self):
        result = _prepare_plots(self.params, npoints=50).load()
        assert result.streamflow.shape == (2, 50)



"""Hydrological modelling framework.

This collection of functions should serve as the main entry point for
hydrological modelling. The entire framework is based on the "model_config"
object. This object is meant to be a container that can be used as needed by
any hydrologic model. For example, it can store datasets directly, paths to
datasets (nc files or other), csv files, basically anything that can be stored
in a dictionary.

It then becomes the user's responsibility to ensure that required data for a
given model be provided in the model_config object both in the data preparation
stage and in the hydrological model implementation. This can be addressed by
a set of pre-defined codes for given model structures. This present package
(hydrological_modelling.py) should contain the logic to:

    1. From the model_config["model_name"] key, select the appropriate function
       (hydrological model) to run.
    2. Pass the model_config object to the correct hydrological modelling
       function.
    3. Parse the model_config object to extract required data for the given
       model, such as: parameters, meteorological data, paths to input files, and catchment characteristics as required
    4. Run the hydrological model with the given data
    5. Return the streamflow (Qsim).

This will make it very easy to add hydrological models, no matter their
complexity. It will also make it easy for newer Python users to implement
models as the logic is clearly exposed and does not make use of complex classes
or other dynamic objects. Models can be added here, and a specific routine
should also be defined to produce the required model_config for each model.

Once this is accomplished, running the model from a model_config object becomes
trivial and allows for easy calibration, regionalisation, analysis and any
other type of interaction.
"""

import numpy as np
import pandas as pd
import xarray as xr

__all__ = ["get_hydrological_model_inputs", "run_hydrological_model"]


def run_hydrological_model(model_config: dict):
    """Hydrological model selector.

    This is the main hydrological model selector. This is the code that looks
    at the "model_config["model_name"]" keyword and calls the appropriate
    hydrological model function.

    Parameters
    ----------
    model_config : dict
        The model configuration object that contains all info to run the model.
        The model function called to run this model shouls always use this object
        and read-in data it requires. It will be up to the user to provide the
        data that the model requires.

    Returns
    -------
    xr.Dataset
        Simulated streamflow from the model, in xarray Dataset format.
    """
    if model_config["model_name"] == "Dummy":
        qsim = _dummy_model(model_config)

    elif model_config["model_name"] == "ADD_OTHER_HERE":
        # ADD OTHER MODELS HERE
        qsim = 0
    else:
        raise NotImplementedError(
            f"The model '{model_config['model_name']}' is not recognized."
        )

    return qsim


def get_hydrological_model_inputs(model_name: str):
    """Required hydrological model inputs for model_config objects.

    Parameters
    ----------
    model_name : str
        Model name that must be one of the models in the list of possible
        models.

    Returns
    -------
    dict
        Elements that must be found in the model_config object.
    """
    if model_name == "Dummy":
        required_config = dict(
            precip="Daily precipitation in mm.",
            temperature="Daily average air temperature in °C",
            drainage_area="Drainage area of the catchment",
            parameters="Model parameters, length 3",
        )

    elif model_name == "ADD_OTHER_HERE":
        # ADD OTHER MODELS HERE
        required_config = {}

    else:
        raise NotImplementedError(f"The model '{model_name}' is not recognized.")

    return required_config


def _dummy_model(model_config: dict):
    """Dummy model.

    Dummy model to show the implementation we should be aiming for. Each model
    will have its own required data that users can pass.

    Parameters
    ----------
    model_config : dict
        The model configuration object that contains all info to run the model.
        The model function called to run this model shouls always use this object
        and read-in data it requires. It will be up to the user to provide the
        data that the model requires.

    Returns
    -------
    xr.Dataset
        Simulated streamflow from the model, in xarray Dataset format.
    """
    # Parse the model_config object to extract required information
    precip = model_config["precip"]
    temperature = model_config["temperature"]
    area = model_config["drainage_area"]
    x = model_config["parameters"]

    # Run the dummy model using these data. Keeping it simple to calculate by
    # hand to ensure the calibration algorithm is working correctly and data
    # are handled correctly
    qsim = np.empty(len(precip))
    for t in range(0, len(precip)):
        qsim[t] = (precip[t] * x[0] + abs(temperature[t]) * x[1]) * x[2] * area

    # For this model, we can convert to xr.dataset by supposing dates. Other
    # models will require some dates in some form (to add to QC checks) in their
    # inputs.
    time = pd.date_range("2024-01-01", periods=len(precip))
    qsim = xr.Dataset(
        data_vars=dict(
            qsim=(["time"], qsim),
        ),
        coords=dict(time=time),
        attrs=dict(description="streamflow simulated by the Dummy Model"),
    )

    return qsim



"""Utilities for testing and releasing xhydro."""

import os
import re
from io import StringIO
from pathlib import Path
from typing import Optional, TextIO, Union

__all__ = [
    "publish_release_notes",
]


def publish_release_notes(
    style: str = "md",
    file: Optional[Union[os.PathLike, StringIO, TextIO]] = None,
    changes: Union[str, os.PathLike] = None,
) -> Optional[str]:
    """Format release history in Markdown or ReStructuredText.

    Parameters
    ----------
    style : {"rst", "md"}
        Use ReStructuredText (`rst`) or Markdown (`md`) formatting. Default: Markdown.
    file : {os.PathLike, StringIO, TextIO, None}
        If provided, prints to the given file-like object. Otherwise, returns a string.
    changes : {str, os.PathLike}, optional
        If provided, manually points to the file where the changelog can be found.
        Assumes a relative path otherwise.

    Returns
    -------
    str or None
        Formatted release notes as a string, if `file` is not provided.

    Notes
    -----
    This function exists solely for development purposes.
    Adapted from xclim.testing.utils.publish_release_notes.
    """
    if isinstance(changes, (str, Path)):
        changes_file = Path(changes).absolute()
    else:
        changes_file = Path(__file__).absolute().parents[2].joinpath("CHANGES.rst")

    if not changes_file.exists():
        raise FileNotFoundError("Changes file not found in xhydro file tree.")

    with open(changes_file) as hf:
        changes = hf.read()

    if style == "rst":
        hyperlink_replacements = {
            r":issue:`([0-9]+)`": r"`GH/\1 <https://github.com/hydrologie/xhydro/issues/\1>`_",
            r":pull:`([0-9]+)`": r"`PR/\1 <https://github.com/hydrologie/xhydro/pull/\>`_",
            r":user:`([a-zA-Z0-9_.-]+)`": r"`@\1 <https://github.com/\1>`_",
        }
    elif style == "md":
        hyperlink_replacements = {
            r":issue:`([0-9]+)`": r"[GH/\1](https://github.com/hydrologie/xhydro/issues/\1)",
            r":pull:`([0-9]+)`": r"[PR/\1](https://github.com/hydrologie/xhydro/pull/\1)",
            r":user:`([a-zA-Z0-9_.-]+)`": r"[@\1](https://github.com/\1)",
        }
    else:
        raise NotImplementedError()

    for search, replacement in hyperlink_replacements.items():
        changes = re.sub(search, replacement, changes)

    if style == "md":
        changes = changes.replace("=========\nChangelog\n=========", "# Changelog")

        titles = {r"\n(.*?)\n([\-]{1,})": "-", r"\n(.*?)\n([\^]{1,})": "^"}
        for title_expression, level in titles.items():
            found = re.findall(title_expression, changes)
            for grouping in found:
                fixed_grouping = (
                    str(grouping[0]).replace("(", r"\(").replace(")", r"\)")
                )
                search = rf"({fixed_grouping})\n([\{level}]{'{' + str(len(grouping[1])) + '}'})"
                replacement = f"{'##' if level=='-' else '###'} {grouping[0]}"
                changes = re.sub(search, replacement, changes)

        link_expressions = r"[\`]{1}([\w\s]+)\s<(.+)>`\_"
        found = re.findall(link_expressions, changes)
        for grouping in found:
            search = rf"`{grouping[0]} <.+>`\_"
            replacement = f"[{str(grouping[0]).strip()}]({grouping[1]})"
            changes = re.sub(search, replacement, changes)

    if not file:
        return changes
    if isinstance(file, (Path, os.PathLike)):
        file = Path(file).open("w")
    print(changes, file=file)



# Created on Fri Dec 8 16:48:34 2023
# @author: Richard Arsenault
"""Calibration package for hydrological models.

This package contains the main framework for hydrological model calibration. It
uses the spotpy calibration package applied on a "model_config" object. This
object is meant to be a container that can be used as needed by any hydrologic
model. For example, it can store datasets directly, paths to datasets (nc files
or other), csv files, basically anything that can be stored in a dictionary.

It then becomes the user's responsibility to ensure that required data for a
given model be provided in the model_config object both in the data preparation
stage and in the hydrological model implementation. This can be addressed by
a set of pre-defined codes for given model structures.

For example, for GR4J, only small datasets are required and can be stored
directly in the model_config dictionary. However, for Hydrotel or Raven models,
maybe it is better to pass paths to netcdf files which can be passed to the
models. This will require pre- and post-processing, but this can easily be
handled at the stage where we create a hydrological model and prepare the data.

The calibration aspect then becomes trivial:

    1. A model_config object is passed to the calibrator.
    2. Lower and upper bounds for calibration parameters are defined and passed
    3. An objective function, optimizer and hyperparameters are also passed.
    4. The calibrator uses this information to develop parameter sets that are
       then passed as inputs to the "model_config" object.
    5. The calibrator launches the desired hydrological model with the
       model_config object (now containing the parameter set) as input.
    6. The appropriate hydrological model function then parses "model_config",
       takes the parameters and required data, launches a simulation and
       returns simulated flow (Qsim).
    7. The calibration package then compares Qobs and Qsim and computes the
       objective function value, and returns this to the sampler that will then
       repeat the process to find optimal parameter sets.
    8. The code returns the best parameter set, objective function value, and
       we also return the simulated streamflow on the calibration period for
       user convenience.

This system has the advantage of being extremely flexible, robust, and
efficient as all data can be either in-memory or only the reference to the
required datasets on disk is passed around the callstack.

Currently, the model_config object has 3 mandatory keywords for the package to
run correctly in all instances:

    - model_config["Qobs"]: Contains the observed streamflow used as the
                            calibration target.

    - model_config["model_name"]: Contains a string referring to the
                                  hydrological model to be run.

    - model_config["parameters"]: While not necessary to provide this, it is
                                  a reserved keyword used by the optimizer.

Any comments are welcome!
"""

from copy import deepcopy

# Import packages
from typing import Optional

import numpy as np
import spotpy
import xarray as xr
from spotpy import analyser
from spotpy.parameter import Uniform

from xhydro.modelling.hydrological_modelling import run_hydrological_model
from xhydro.modelling.obj_funcs import (
    _get_objfun_minimize_or_maximize,
    _get_optimizer_minimize_or_maximize,
    get_objective_function,
)

__all__ = ["perform_calibration"]


class SpotSetup:
    """Create the spotpy calibration system that is used for hydrological model calibration.

    Parameters
    ----------
    model_config : dict
        The model configuration object that contains all info to run the model.
        The model function called to run this model should always use this object and read-in data it requires.
        It will be up to the user to provide the data that the model requires.
    bounds_high : np.array
        High bounds for the model parameters to be calibrated. Spotpy will sample parameter sets from
        within these bounds. The size must be equal to the number of parameters to calibrate.
    bounds_low : np.array
        Low bounds for the model parameters to be calibrated. Spotpy will sample parameter sets from
        within these bounds. The size must be equal to the number of parameters to calibrate.
    obj_func : str
        The objective function used for calibrating. Can be any one of these:

            - "abs_bias" : Absolute value of the "bias" metric
            - "abs_pbias": Absolute value of the "pbias" metric
            - "abs_volume_error" : Absolute value of the volume_error metric
            - "agreement_index": Index of agreement
            - "correlation_coeff": Correlation coefficient
            - "kge" : Kling Gupta Efficiency metric (2009 version)
            - "kge_mod" : Kling Gupta Efficiency metric (2012 version)
            - "mae": Mean Absolute Error metric
            - "mare": Mean Absolute Relative Error metric
            - "mse" : Mean Square Error metric
            - "nse": Nash-Sutcliffe Efficiency metric
            - "r2" : r-squared, i.e. square of correlation_coeff.
            - "rmse" : Root Mean Square Error
            - "rrmse" : Relative Root Mean Square Error (RMSE-to-mean ratio)
            - "rsr" : Ratio of RMSE to standard deviation.

    take_negative : bool
        Inidactor to take the negative of the objective function value in optimization to ensure convergence
        in the right direction.
    mask : np.array
        A vector indicating which values to preserve/remove from the objective function computation. 0=remove, 1=preserve.
    transform : str
        The method to transform streamflow prior to computing the objective function. Can be one of:
        Square root ('sqrt'), inverse ('inv'), or logarithmic ('log') transformation.
    epsilon : float
        Used to add a small delta to observations for log and inverse transforms, to eliminate errors
        caused by zero flow days (1/0 and log(0)). The added perturbation is equal to the mean observed streamflow
        times this value of epsilon.

    Returns
    -------
    SpotSetup object for calibration
    """

    def __init__(
        self,
        model_config: dict,
        bounds_high: np.array,
        bounds_low: np.array,
        obj_func: str = None,
        take_negative: bool = False,
        mask: np.array = None,
        transform: str = None,
        epsilon: float = 0.01,
    ):
        """
        Initialize the SpotSetup object.

        The initialization of the SpotSetup object includes a generic
        "model_config" object containing hydrological modelling data required,
        low and high parameter bounds, as well as an objective function.
        Depending on the objective function, either spotpy or hydroeval will
        compute the value, since some functions are found only in one package.

        Parameters
        ----------
        model_config : dict
            The model configuration object that contains all info to run the model.
            The model function called to run this model should always use this object and read-in data it requires.
            It will be up to the user to provide the data that the model requires.
        obj_func : str
            The objective function used for calibrating. Can be any one of these:

                - "abs_bias" : Absolute value of the "bias" metric
                - "abs_pbias": Absolute value of the "pbias" metric
                - "abs_volume_error" : Absolute value of the volume_error metric
                - "agreement_index": Index of agreement
                - "correlation_coeff": Correlation coefficient
                - "kge" : Kling Gupta Efficiency metric (2009 version)
                - "kge_mod" : Kling Gupta Efficiency metric (2012 version)
                - "mae": Mean Absolute Error metric
                - "mare": Mean Absolute Relative Error metric
                - "mse" : Mean Square Error metric
                - "nse": Nash-Sutcliffe Efficiency metric
                - "r2" : r-squared, i.e. square of correlation_coeff.
                - "rmse" : Root Mean Square Error
                - "rrmse" : Relative Root Mean Square Error (RMSE-to-mean ratio)
                - "rsr" : Ratio of RMSE to standard deviation.
        bounds_high : np.array
            High bounds for the model parameters to be calibrated. Spotpy will sample parameter sets from
            within these bounds. The size must be equal to the number of parameters to calibrate.
        bounds_low : np.array
            Low bounds for the model parameters to be calibrated. Spotpy will sample parameter sets from
            within these bounds. The size must be equal to the number of parameters to calibrate.
        evaluations : int
            Maximum number of model evaluations (calibration budget) to perform before stopping the calibration process.
        algorithm : str
            The optimization algorithm to use. Currently, "DDS" and "SCEUA" are available, but more can be easily added.
        take_negative : bool
            Wether to take the negative of the objective function value in optimization to ensure convergence
            in the right direction.
        mask : np.array, optional
            A vector indicating which values to preserve/remove from the objective function computation. 0=remove, 1=preserve.
        transform : str, optional
            The method to transform streamflow prior to computing the objective function. Can be one of:
            Square root ('sqrt'), inverse ('inv'), or logarithmic ('log') transformation.
        epsilon : float
            Used to add a small delta to observations for log and inverse transforms, to eliminate errors
            caused by zero flow days (1/0 and log(0)). The added perturbation is equal to the mean observed streamflow
            times this value of epsilon.

        Returns
        -------
        SpotSetup object for calibration.
        """
        # Gather the model_config dictionary and obj_func string, and other
        # optional arguments.
        self.model_config = deepcopy(model_config)
        self.obj_func = obj_func
        self.mask = mask
        self.transform = transform
        self.epsilon = epsilon
        self.take_negative = take_negative

        # Create the sampler for each parameter based on the bounds
        self.parameters = [
            Uniform("param" + str(i), bounds_low[i], bounds_high[i])
            for i in range(0, len(bounds_high))
        ]

    def simulation(self, x):
        """Simulation function for spotpy.

        This is where the optimizer generates a parameter set from within the
        given bounds and generates the simulation results. We add the parameter
        "x" that is generated by spotpy to the model_config object at the
        reserved keyword "parameters".

        Parameters
        ----------
        x : array_like
            Tested parameter set.

        Returns
        -------
        array_like
            Simulated streamflow.
        """
        self.model_config.update({"parameters": x})

        # Run the model and return qsim, with model_config containing the
        # tested parameter set.
        qsim = run_hydrological_model(self.model_config)

        # Return the array of values from qsim for the objective function eval.
        return qsim["qsim"].values

    def evaluation(self):
        """Evaluation function for spotpy.

        Here is where we get the observed streamflow and make it available to
        compare the simulation and compute an objective function. It has to be
        in the Qobs keyword, although with some small changes
        model_config['Qobs'] could be a string to a file. Probably more
        efficient to load it into memory during preprocessing anyway to
        prevent recurring input/output and associated overhead. Currently, the
        package supposes that Qobs and Qsim have the same length, but this can
        be changed in the model_config parameterization and adding conditions
        here.

        Returns
        -------
        array_like
            Observed streamflow.
        """
        qobs = self.model_config["qobs"]
        if isinstance(qobs, xr.Dataset):
            qobs = qobs["qobs"]

        return qobs

    def objectivefunction(
        self,
        simulation,
        evaluation,
    ):
        """
        Objective function for spotpy.

        This function is where spotpy computes the objective function.

        Parameters
        ----------
        simulation : array_like
            Vector of simulated streamflow.
        evaluation : array_like
            Vector of observed streamflow to compute objective function.

        Returns
        -------
        float
            The value of the objective function.
        """
        obj_fun_val = get_objective_function(
            evaluation,
            simulation,
            obj_func=self.obj_func,
            take_negative=self.take_negative,
            mask=self.mask,
            transform=self.transform,
            epsilon=self.epsilon,
        )

        return obj_fun_val


def perform_calibration(
    model_config: dict,
    obj_func: str,
    bounds_high: np.array,
    bounds_low: np.array,
    evaluations: int,
    algorithm: str = "DDS",
    mask: Optional[np.array] = None,
    transform: Optional[str] = None,
    epsilon: float = 0.01,
    sampler_kwargs: Optional[dict] = None,
):
    """Perform calibration using spotpy.

    This is the entrypoint for the model calibration. After setting-up the
    model_config object and other arguments, calling "perform_calibration" will
    return the optimal parameter set, objective function value and simulated
    flows on the calibration period.

    Parameters
    ----------
    model_config : dict
        The model configuration object that contains all info to run the model.
        The model function called to run this model should always use this object and read-in data it requires.
        It will be up to the user to provide the data that the model requires.
    obj_func : str
        The objective function used for calibrating. Can be any one of these:

            - "abs_bias" : Absolute value of the "bias" metric
            - "abs_pbias": Absolute value of the "pbias" metric
            - "abs_volume_error" : Absolute value of the volume_error metric
            - "agreement_index": Index of agreement
            - "correlation_coeff": Correlation coefficient
            - "kge" : Kling Gupta Efficiency metric (2009 version)
            - "kge_mod" : Kling Gupta Efficiency metric (2012 version)
            - "mae": Mean Absolute Error metric
            - "mare": Mean Absolute Relative Error metric
            - "mse" : Mean Square Error metric
            - "nse": Nash-Sutcliffe Efficiency metric
            - "r2" : r-squared, i.e. square of correlation_coeff.
            - "rmse" : Root Mean Square Error
            - "rrmse" : Relative Root Mean Square Error (RMSE-to-mean ratio)
            - "rsr" : Ratio of RMSE to standard deviation.

    bounds_high : np.array
        High bounds for the model parameters to be calibrated. Spotpy will sample parameter sets from
        within these bounds. The size must be equal to the number of parameters to calibrate.
    bounds_low : np.array
        Low bounds for the model parameters to be calibrated. Spotpy will sample parameter sets from
        within these bounds. The size must be equal to the number of parameters to calibrate.
    evaluations : int
        Maximum number of model evaluations (calibration budget) to perform before stopping the calibration process.
    algorithm : str
        The optimization algorithm to use. Currently, "DDS" and "SCEUA" are available, but more can be easily added.
    mask : np.array, optional
        A vector indicating which values to preserve/remove from the objective function computation. 0=remove, 1=preserve.
    transform : str, optional
        The method to transform streamflow prior to computing the objective function. Can be one of:
        Square root ('sqrt'), inverse ('inv'), or logarithmic ('log') transformation.
    epsilon : scalar float
        Used to add a small delta to observations for log and inverse transforms, to eliminate errors
        caused by zero flow days (1/0 and log(0)). The added perturbation is equal to the mean observed streamflow
        times this value of epsilon.
    sampler_kwargs : dict
        Contains the keywords and hyperparameter values for the optimization algorithm.
        Keywords depend on the algorithm choice. Currently, SCEUA and DDS are supported with
        the following default values:
        - SCEUA: dict(ngs=7, kstop=3, peps=0.1, pcento=0.1)
        - DDS: dict(trials=1)

    Returns
    -------
    best_parameters : array_like
        The optimized parameter set.
    qsim : xr.Dataset
        Simulated streamflow using the optimized parameter set.
    bestobjf : float
        The best objective function value.
    """
    # Get objective function and algo optimal convergence direction. Necessary
    # to ensure that the algorithm is optimizing in the correct direction
    # (maximizing or minimizing). This code determines the required direction
    # for the objective function and the working direction of the algorithm.
    of_maximize = _get_objfun_minimize_or_maximize(obj_func)
    algo_maximize = _get_optimizer_minimize_or_maximize(algorithm)

    # They are not working in the same direction. Take the negative of the OF.
    if of_maximize != algo_maximize:
        take_negative = True
    else:
        take_negative = False

    # Set up the spotpy object to prepare the calibration
    spotpy_setup = SpotSetup(
        model_config,
        bounds_high=bounds_high,
        bounds_low=bounds_low,
        obj_func=obj_func,
        take_negative=take_negative,
        mask=mask,
        transform=transform,
        epsilon=epsilon,
    )

    # Select an optimization algorithm and parameterize it, then run the
    # optimization process.
    sampler_kwargs = deepcopy(sampler_kwargs) or None
    if sampler_kwargs is None:
        sampler_kwargs = {}

    if algorithm == "DDS":
        sampler = spotpy.algorithms.dds(
            spotpy_setup, dbname="DDS_optim", dbformat="ram", save_sim=False
        )

        # Get the sampler hyperparameters, either default or user-provided.
        defaults = {"trials": 1}
        sampler_kwargs = defaults | sampler_kwargs

        # Ensure there is only 1 hyperparameter passed by the user, if
        # applicable
        if len(sampler_kwargs) == 1:
            sampler.sample(evaluations, **sampler_kwargs)
        else:
            raise ValueError(
                "sampler_kwargs should only contain the keyword 'trials' when using DDS."
            )

    elif algorithm == "SCEUA":
        sampler = spotpy.algorithms.sceua(
            spotpy_setup, dbname="SCEUA_optim", dbformat="ram", save_sim=False
        )
        # Get the sampler hyperparameters, either default or user-provided.
        defaults = {"ngs": 7, "kstop": 3, "peps": 0.1, "pcento": 0.1}
        sampler_kwargs = defaults | sampler_kwargs

        # If the user provided a custom sampler hyperparameter set.
        if len(sampler_kwargs) == 4:
            sampler.sample(evaluations, **sampler_kwargs)
        else:
            raise ValueError(
                "sampler_kwargs should only contain the keywords [ngs, kstop, peps, pcento] when using SCEUA."
            )

    # Gather optimization results
    results = sampler.getdata()

    # Get the best parameter set
    best_parameters = analyser.get_best_parameterset(
        results, like_index=1, maximize=of_maximize
    )
    best_parameters = [best_parameters[0][i] for i in range(0, len(best_parameters[0]))]

    # Get the best objective function as well depending if maximized or
    # minimized
    if of_maximize:
        _, bestobjf = analyser.get_maxlikeindex(results)
    else:
        _, bestobjf = analyser.get_minlikeindex(results)

    # Reconvert objective function if required.
    if take_negative:
        bestobjf = bestobjf * -1

    # Update the parameter set to put the best parameters in model_config...
    model_config = deepcopy(model_config)
    model_config.update({"parameters": best_parameters})

    # ... which can be used to run the hydrological model and get the best Qsim.
    qsim = run_hydrological_model(model_config)

    # Return the best parameters, qsim and best objective function value.
    return best_parameters, qsim, bestobjf



"""Test suite for hydrological modelling in hydrological_modelling.py"""

import numpy as np
import pytest

from xhydro.modelling.hydrological_modelling import (
    get_hydrological_model_inputs,
    run_hydrological_model,
)


def test_hydrological_modelling():
    """Test the hydrological models as they become online"""
    # Test the dummy model
    model_config = {
        "precip": np.array([10, 11, 12, 13, 14, 15]),
        "temperature": np.array([10, 3, -5, 1, 15, 0]),
        "qobs": np.array([120, 130, 140, 150, 160, 170]),
        "drainage_area": np.array([10]),
        "model_name": "Dummy",
        "parameters": np.array([5, 5, 5]),
    }
    qsim = run_hydrological_model(model_config)
    assert qsim["qsim"].values[3] == 3500.00

    # Test the exceptions for new models
    model_config.update(model_name="ADD_OTHER_HERE")
    qsim = run_hydrological_model(model_config)
    assert qsim == 0


def test_import_unknown_model():
    """Test for unknown model"""
    with pytest.raises(NotImplementedError) as pytest_wrapped_e:
        model_config = {"model_name": "fake_model"}
        _ = run_hydrological_model(model_config)
        assert pytest_wrapped_e.type == NotImplementedError


def test_get_unknown_model_requirements():
    """Test for required inputs for models with unknown name"""
    with pytest.raises(NotImplementedError) as pytest_wrapped_e:
        model_name = "fake_model"
        _ = get_hydrological_model_inputs(model_name)
        assert pytest_wrapped_e.type == NotImplementedError


def test_get_model_requirements():
    """Test for required inputs for models"""
    model_name = "Dummy"
    required_config = get_hydrological_model_inputs(model_name)
    print(required_config.keys())
    assert len(required_config.keys()) == 4
